//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2022
//#
//#
//# File Name:  active_transfer_spi.v
//# Author:     R. Jolly
//# Date:       November 13, 2022
//# Revision:   A
//#
//# Development: USB to Slave SPI interface 
//# Application: Altera Cyclone 10 FPGA
//# Description: This file contains verilog code which will allow access
//#              to an eight bit bus in the FPGA. The FPGA receives its commands via
//#				 USB.
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			11/13/22 	A			Created			RJJ
//#
//#									
//#
//#######################################################################

`include "../src/define.v"

`timescale 1 ns / 10 ps

module active_transfer_spi
  (
	input wire                 CLK,
	input wire                 RST_N,

  
	output wire                SCLK, //To external 
    input wire                 MISO, //From external
	output wire                MOSI,
	inout  wire                MIOSIO_1,
	output wire                SS,
  
	output reg                 USB_RXF_N,
	output reg                 USB_TXE_N,
    inout wire  [7:0]          USB_DATA,

	input  wire                USB_RD_N,
	input  wire                USB_WR,
	input  wire                USB_SI,
	
	output reg [3:0]           STATE_OUT,
	output wire [7:0]          TEST_OUT


	);



   //----------------------------------------------
   // Parameter Declarations                        
   //----------------------------------------------
   parameter IDLE                    = 0,
              RECEIVE_DATA           = 1,
              STORE_BYTE             = 2,
              SEND_TO_HOST           = 3,
              RECEIVE_COMPLETE       = 4,
              SEND_CMD_BYTE          = 5,
			  SEND_CMD_COMPLETE      = 6,
              CHECK_FOR_BYTE         = 7,
              READ_CHECK_BYTE        = 8;
		 
   parameter  IDLE_TR                = 0,
              WRITE_SEND_CMD         = 1,
			  WRITE_CMD_COMPLETE     = 2,
              WRITE_REGISTER         = 3,
              LOAD_SHIFT_REG         = 4,
              DELAY_WRITE_DATA_BYTE  = 5,
              WRITE_DATA_BYTE        = 6,
              WRITE_DATA_COMPLETE	 = 7;
		 
   parameter READ_CHECK_INVALID      = 8'hFF;
   parameter NUMBER_BYTES_TO_READ    = 2;
   
   //SPI Host Command Processor
   parameter  HOST_CMD_STATE_IDLE    = 0,
              HOST_CMD_CHECK_IN_BYTE = 1,
              HOST_CMD_WRITE_BYTE    = 2,
              HOST_CMD_READ_BYTE     = 3,
			  HOST_WRITE_DELAY       = 4,
			  HOST_WRITE_COMPLETE    = 5;
			  
   //FT220X Command Bytes
   parameter FT220X_READ_COMMAND     = 8'h02;  
   parameter FT220X_WRITE_COMMAND    = 8'h00;
   parameter FT220X_DEFAULT_COMMAND  = 8'hFF; 
   parameter FT220X_STATUS_COMMAND   = 8'h35;   
   parameter CMD                     = 2'b01;
   parameter READ                    = 2'b11;
   parameter WRITE                   = 2'b10;
   
   //SPI Write Delay Count for spi_master_we_in
   parameter SPI_WRITE_DELAY_COUNT   = 22;
   
   //Automatic Host Read Timer
`ifdef SIM
   parameter HOST_AUTO_READ_TIMER_MAX = 32'hFFFFFFFF;
   parameter HOST_AUTO_READ_TIMER_MIN = 32'hFFFF9E58;
   parameter HOST_AUTO_READ_INITIATE  = 32'hFFFF9E5F;  
`else
   parameter HOST_AUTO_READ_TIMER_MAX = 32'hFFFFFFFF;
   parameter HOST_AUTO_READ_TIMER_MIN = 32'hFFF85ED8;
   parameter HOST_AUTO_READ_INITIATE  = 32'hFFF85EDF; /*10 millisecond interval */
`endif
   
		 

   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
   	reg [8:0]                 state_receive, next_receive;
   	reg [7:0]                 state_transmit, next_transmit;

	// Storage Registers
	reg [7:0]                  rcv_storage_reg;
	reg [2:0]                  store_reg_state; 
	
	
	//SPI Bus control registers
    wire                      spi_miso;
    wire                      spi_clk;
	wire                      spi_mosi;
    wire                      spi_cs;
	reg                       ss_reg;
	
	//spi_iface_master control registers
	reg  [7:0]                 spi_master_din;
    reg                        start_spi_write;
	reg                        spi_master_we_in;
	wire                       spi_master_rdy_in;
	wire [7:0]                 spi_master_dout;
	wire                       spi_master_we_out_n;
	reg                        spi_master_rdy_out;
	
	reg  [7:0]                 spi_delay_counter;
	reg                        spi_write_delay_flag;
	
	//USB Interface Registers
	reg  [7:0]                 tx_data_reg;
	reg  [7:0]                 rx_byte_present;
	
	//SPI Read state machine Registers
    reg                        spi_host_read;
	reg  [31:0]                spi_host_read_counter;
    reg                        spi_check_for_byte;
    reg                        spi_host_check_byte_valid;
    reg                        spi_read_data;
	reg  [7:0]                 spi_slave_number_of_bytes;
	
	//SPI Host Command Processor
	reg  [1:0]                 spi_cmd_byte_selection;
	reg  [2:0]                 host_cmd_state;
	
	//MIOSIO Direction control registers
	reg                        write_to_miosio;
	reg                        write_to_miosio_reg;
	wire                       miosio_miso;



	
   	
`ifdef SIM
   reg [8*26:1] state_name_receive;
   reg [8*26:1] state_name_transmit;
`endif

   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------
   assign USB_DATA = state_receive[RECEIVE_COMPLETE] ? rcv_storage_reg : 8'hz;
   //assign USB_RD_N = state[WRITE_REGISTER] ? 1'b0 : 1'b1;
   //assign tx_data_reg = state[SEND_TO_HOST] ? 8'hz : (automated_write ? automated_write_byte : USB_DATA);
   //assign write_tx_byte = state[WRITE_DATA_BYTE] ? 1'b0: 1'b1;

   assign    MOSI = spi_mosi;
   assign    spi_miso = MISO;
   assign    SCLK = spi_clk;
   assign    SS = ss_reg;
   assign    miosio_miso = MIOSIO_1;
   assign    MIOSIO_1 = write_to_miosio ? spi_mosi : 1'bz;
   
	
   //TEST Only
   assign TEST_OUT[0] = 1'b0;
   assign TEST_OUT[1] = spi_host_check_byte_valid;
   assign TEST_OUT[2] = 1'b0;
   assign TEST_OUT[3] = 1'b0;
   assign TEST_OUT[4] = 1'b0;
   assign TEST_OUT[5] = 1'b0;
   assign TEST_OUT[6] = 1'b0;
   assign TEST_OUT[7] = 1'b0;   
	
   //-----------------------------------------------
   // Create a State Test Output Signal
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  STATE_OUT <= 0;
	else
	  begin
	     if (state_receive[IDLE])
           STATE_OUT <= 4'h0;
	     if (state_receive[RECEIVE_DATA])
	       STATE_OUT <= 4'h1;
	     if (state_receive[STORE_BYTE])
	       STATE_OUT <= 4'h2;
	     if (state_receive[SEND_TO_HOST])
	       STATE_OUT <= 4'h3;
	     if (state_receive[RECEIVE_COMPLETE])
	       STATE_OUT <= 4'h4;
	     if (state_receive[SEND_CMD_BYTE])
	       STATE_OUT <= 4'h5;
	     if (state_receive[SEND_CMD_COMPLETE])
	       STATE_OUT <= 4'h6;
	     if (state_receive[CHECK_FOR_BYTE])
	       STATE_OUT <= 4'h7;
	     if (state_receive[READ_CHECK_BYTE])
	       STATE_OUT <= 4'h8;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Register to control the direction of the MIOSIO_1 
   // signal. During the Write cycle, the Master SPI
   // will write to the Slave. During Read cycles,
   // the Master SPI will read from Slave.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
   begin
	 if (!RST_N)
	 begin
		write_to_miosio <= 1'b0;
		write_to_miosio_reg <= 1'b0;
	 end
	 else
	 begin
        if(!ss_reg & !write_to_miosio_reg)
        begin
		   write_to_miosio <= 1'b1;
		   write_to_miosio_reg <= 1'b1;
		end
		else if(state_receive[CHECK_FOR_BYTE] & write_to_miosio_reg)
        begin
		   write_to_miosio <= 1'b0;
		   write_to_miosio_reg <= 1'b1;
		end
		else if(state_transmit[WRITE_SEND_CMD] & write_to_miosio_reg)
        begin
		   write_to_miosio <= 1'b0;
		   write_to_miosio_reg <= 1'b1;
		end
        else if(ss_reg)
        begin
		   write_to_miosio <= 1'b0;
		   write_to_miosio_reg <= 1'b0;
		end
	 end
   end
   
   //-----------------------------------------------
   //Transmit: Send the SS based on spi_cs and the 
   // host_cmd_state. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     ss_reg <= 1'b1;
	else
	  begin
	     if(!spi_cs & ((host_cmd_state == HOST_CMD_CHECK_IN_BYTE) | (host_cmd_state == HOST_CMD_WRITE_BYTE)))
	         ss_reg <= 1'b0;
		 else if(spi_cs & (state_receive[IDLE] & state_transmit[IDLE]))
	         ss_reg <= 1'b1;
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   //Transmit: Write to the SPI only based on the 
   // state of the host_cmd_state. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     start_spi_write <= 1;
	else
	  begin
	     if (host_cmd_state == HOST_CMD_WRITE_BYTE)
	         start_spi_write <= 0;
		 else
	         start_spi_write <= 1;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   //Transmit: Store the USB_Data byte into the UART
   // transmist register. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if(!RST_N)
	begin
	     //start_spi_write <= 1;
		 tx_data_reg <= 0;
	end
	else
	  begin
	     if(state_transmit[LOAD_SHIFT_REG])
		 begin
	         //start_spi_write <= 0;
			 tx_data_reg <= USB_DATA;
	     end
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Transmit: write_tx_byte Signal 
   //-----------------------------------------------
	always @(posedge CLK or negedge RST_N)
	begin
        if (!RST_N)
		begin
	       spi_master_din <= 0;
	       spi_master_we_in <= 1'b0;
	       spi_delay_counter <= 0;
	       spi_write_delay_flag <= 1'b0;
		end
		else 
		begin
			if(state_transmit[DELAY_WRITE_DATA_BYTE]  & !spi_write_delay_flag)
			begin
		       spi_master_din <= tx_data_reg;
	           spi_master_we_in <= 1'b1;
               spi_write_delay_flag <= 1'b1;
			end
			else if(state_transmit[WRITE_SEND_CMD] & !spi_write_delay_flag)
			begin
			   case(spi_cmd_byte_selection)
			   CMD:
			   begin
		          spi_master_din <= FT220X_STATUS_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   READ:
			   begin
		          spi_master_din <= FT220X_READ_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   WRITE:
			   begin
		          spi_master_din <= FT220X_WRITE_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   default:
			   begin
		          spi_master_din <= FT220X_DEFAULT_COMMAND;
	              spi_master_we_in <= 1'b0;
			   end
			   endcase
	           spi_master_we_in <= 1'b1;
               spi_write_delay_flag <= 1'b1;
			end
			else if((state_transmit[WRITE_DATA_BYTE]  | state_transmit[WRITE_CMD_COMPLETE]) & spi_write_delay_flag)
			begin
	           //spi_master_we_in <= 1'b0;
	           
			   
			   if(spi_delay_counter < SPI_WRITE_DELAY_COUNT)
		       begin
		          spi_delay_counter <= spi_delay_counter + 1'd1;
		          if(spi_delay_counter > 6'h6)
		            spi_master_we_in <= 1'b0;
		      end
			end
			else if((!spi_master_rdy_in) & (state_transmit[LOAD_SHIFT_REG] | state_transmit[WRITE_DATA_COMPLETE]))
			begin
	           spi_master_we_in <= 1'b0;
		       spi_delay_counter <= 0;
			   spi_write_delay_flag <= 1'b0;
		    end
			
			//Byte Transmit For Read Command
			if(state_receive[SEND_CMD_BYTE])
			begin
			   case(spi_cmd_byte_selection)
			   CMD:
			   begin
		          spi_master_din <= FT220X_STATUS_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   READ:
			   begin
		          spi_master_din <= FT220X_READ_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   WRITE:
			   begin
		          spi_master_din <= FT220X_WRITE_COMMAND;
	              spi_master_we_in <= 1'b1;
			   end
			   default:
			   begin
		          spi_master_din <= FT220X_DEFAULT_COMMAND;
	              spi_master_we_in <= 1'b0;
			   end
			   endcase
			end
			else if(state_receive[SEND_CMD_COMPLETE])
               spi_master_we_in <= 1'b0;
		end	
	end // always @ (posedge CLK or negedge RESET)   


   //-----------------------------------------------
   //Transmit: Assert the USB_TXE_N signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     USB_TXE_N <= 1'b1;
	else
	  begin
	     if(state_transmit[WRITE_REGISTER])
	         USB_TXE_N <= 1'b0;
		 else if(state_transmit[WRITE_DATA_COMPLETE])
	         USB_TXE_N <= 1'b1;
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Receive: spi_master_rdy_out register reads the available byte
   // from the Slave SPI receive FIFO. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	     spi_master_rdy_out <= 1'b0;
	end
	else
	  begin
	     if (state_receive[RECEIVE_DATA] | state_receive[CHECK_FOR_BYTE])
		 begin
	         spi_master_rdy_out <= 1'b1;
		 end
		 else
		 begin
	         spi_master_rdy_out <= 1'b0;
		 end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Receive: spi_master_we_out_n register indicates
   // whether the SPI Unit has completed the read
   // cycle from the Slave SPI receive FIFO. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
		 spi_read_data <= 1'b0;
		 rcv_storage_reg <= 0;
	end
	else
	  begin
	     if (state_receive[STORE_BYTE])
		 begin
	         spi_read_data <= 1'b1;
		 end
		 else if(!spi_master_we_out_n & spi_read_data)
		 begin
	         spi_read_data <= 1'b0;
			 rcv_storage_reg <= spi_master_dout;
		 end
		 else if(spi_master_we_out_n & !spi_read_data)
		 begin
		    //
		 end
		 else if(spi_host_check_byte_valid)
		 begin
            rcv_storage_reg <= spi_master_dout;
		 end
		 
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Receive: USB_RXF_N output register based on  
   // state[RECEIVE_COMPLETE]. Indicates to User Code
   // that a byte is ready to be read from the Receive
   // Unit.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     USB_RXF_N <= 1'b1;
	else
	  begin
	     if(state_receive[SEND_TO_HOST] & USB_RD_N)
	         USB_RXF_N <= 1'b0;
		 else if(state_receive[SEND_TO_HOST] & !USB_RD_N)
	         USB_RXF_N <= 1'b1;
		 else if(state_receive[IDLE])
	         USB_RXF_N <= 1'b1;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Host Command Processor:  State Machine to determine 
   // which command to send to the Host FT220X chip. 
   // 
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	     spi_cmd_byte_selection <= 2'b00;
		 host_cmd_state <= 3'b000; //HOST_CMD_STATE_IDLE;
    end
	else
	  begin
	     case(host_cmd_state)
		 HOST_CMD_STATE_IDLE:
		 begin
	        if(spi_check_for_byte)
			begin
	           spi_cmd_byte_selection <= READ;
		       host_cmd_state <= HOST_CMD_CHECK_IN_BYTE;
			end
		    else if(!USB_WR)
			begin
	           spi_cmd_byte_selection <= WRITE;
		       host_cmd_state <= HOST_CMD_WRITE_BYTE;
			end
 		 end
		 HOST_CMD_CHECK_IN_BYTE:
		 begin
	        if(spi_master_rdy_in)
			begin
		       host_cmd_state <= HOST_CMD_READ_BYTE;
			end
			else
			   host_cmd_state <= HOST_CMD_CHECK_IN_BYTE;
 		 end
		 HOST_CMD_WRITE_BYTE:
		 begin
	        if(spi_master_rdy_in)
			begin
		       host_cmd_state <= HOST_WRITE_DELAY;
			end
			else
			   host_cmd_state <= HOST_CMD_WRITE_BYTE;
 		 end
		 HOST_CMD_READ_BYTE:
		 begin
	        if(spi_master_rdy_out)
			begin
               spi_cmd_byte_selection <= READ;
		       host_cmd_state <= HOST_CMD_STATE_IDLE;
			end
			else
			   host_cmd_state <= HOST_CMD_READ_BYTE;
 		 end
		 HOST_WRITE_DELAY:
		 begin
	        if(state_transmit[WRITE_DATA_COMPLETE])
			begin
		       host_cmd_state <= HOST_WRITE_COMPLETE;
			end
			else
			   host_cmd_state <= HOST_WRITE_DELAY;
 		 end
		 HOST_WRITE_COMPLETE:
		 begin
	        if(USB_WR)
			begin
		       host_cmd_state <= HOST_CMD_STATE_IDLE;
			end
			else
			   host_cmd_state <= HOST_WRITE_COMPLETE;
 		 end
		 default:
		 begin
		    if(spi_master_rdy_out)
			begin
               spi_cmd_byte_selection <= READ;
		       host_cmd_state <= HOST_CMD_STATE_IDLE;
			end
 		 end
		 endcase
	  end
     end // always @ (posedge CLK or negedge RST_N)

	 //-----------------------------------------------
   // Receive: State Machine to periodically read  
   // the Slave SPI device and determine if there
   // are bytes waiting to be read. 
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
    begin
      if (!RST_N)
      begin
          spi_check_for_byte <= 1'b0;
		  spi_host_read_counter <= HOST_AUTO_READ_TIMER_MAX;
		  spi_host_check_byte_valid = 1'b0;
      end
      else
      begin
         if(spi_host_read_counter > HOST_AUTO_READ_TIMER_MIN)
         begin
            spi_host_read_counter <= spi_host_read_counter - 1'd1;
            if(spi_host_read_counter == HOST_AUTO_READ_INITIATE)
			begin
               spi_check_for_byte <= 1'b1;
			   spi_host_check_byte_valid = 1'b0;
			end
         end
         else
		 begin
		    spi_check_for_byte <= 1'b0;
            spi_host_read_counter <= HOST_AUTO_READ_TIMER_MAX;
         end 
		 
		 //Check Byte to determine if data is available in FIFO
		 if(state_receive[CHECK_FOR_BYTE])
         begin
		    if(!MISO/*spi_master_dout != READ_CHECK_INVALID*/)
			   spi_host_check_byte_valid = 1'b1;
         end
		 else if(state_receive[IDLE])
			   spi_host_check_byte_valid = 1'b0;	 
      end
	end

   //-----------------------------------------------
   // Receive: Read Bytes from the Slave SPI
   // once it has been determined that the Slave SPI
   // FIFO has legitimate bytes in queue.
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
    begin
      if (!RST_N)
      begin
          spi_host_read <= 1'b0;
		  spi_slave_number_of_bytes <= NUMBER_BYTES_TO_READ;
      end
      else
      begin
	    if(spi_host_check_byte_valid)
		   spi_host_read <= 1'b0;
		else if(state_receive[STORE_BYTE])
		   spi_host_read <= 1'b0;
	  end
	end
	
   //--------------------------------------------------
   // SPI Read State Machine Receive
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state_receive <= 9'h00;
	     state_receive[IDLE] <= 1'b1;
	  end
	else
	  state_receive <= next_receive;
     end

   always @ ( /*AUTOSENSE*/state_receive  
		 or spi_check_for_byte or spi_host_read or rx_byte_present
		 or USB_DATA or USB_RD_N or USB_RXF_N or USB_TXE_N
		 or spi_master_rdy_out or spi_master_we_out_n or spi_master_we_in
		 or spi_master_rdy_in)
   begin
	    next_receive = 9'h00;

	if (state_receive[IDLE])
	begin
        if (spi_host_read)
			next_receive[RECEIVE_DATA] = 1'b1;
        else if (spi_check_for_byte)
			next_receive[SEND_CMD_BYTE] = 1'b1;
		else
			next_receive[IDLE] = 1'b1;
	end

	if (state_receive[RECEIVE_DATA])
	       next_receive[STORE_BYTE] = 1'b1;

	  if(state_receive[STORE_BYTE]) 
	  begin
	     if (!spi_host_read)
	       next_receive[SEND_TO_HOST] = 1'b1;
	     else
	         next_receive[STORE_BYTE] = 1'b1;
	  end 	     

	if (state_receive[SEND_TO_HOST]) 
	  begin
	     if (!USB_RXF_N)
			  next_receive[RECEIVE_COMPLETE] = 1'b1;
	     else
	         next_receive[SEND_TO_HOST] = 1'b1;
	  end 	     

	  if (state_receive[RECEIVE_COMPLETE]) 
	  begin
	     if (!USB_RD_N)
			  next_receive[IDLE] = 1'b1;
	     else
	         next_receive[RECEIVE_COMPLETE] = 1'b1;
	  end 	     

	  if (state_receive[SEND_CMD_BYTE])
	     if (spi_master_rdy_in)
			  next_receive[SEND_CMD_COMPLETE] = 1'b1;
	     else
	       next_receive[SEND_CMD_BYTE] = 1'b1;
		   
	  if(state_receive[SEND_CMD_COMPLETE])
	  begin
	     if(!spi_master_rdy_in)
		    next_receive[CHECK_FOR_BYTE] = 1'b1;
	     else
		    next_receive[SEND_CMD_COMPLETE] = 1'b1;
	  end

	  if (state_receive[CHECK_FOR_BYTE])
	  begin
	     if(!spi_master_we_out_n)
	       next_receive[READ_CHECK_BYTE] = 1'b1;
	     else
	         next_receive[CHECK_FOR_BYTE] = 1'b1;
      end
	  
	  if (state_receive[READ_CHECK_BYTE]) 
	  begin
	     if (spi_master_we_out_n)
            next_receive[IDLE] = 1'b1;
		 else if(spi_host_check_byte_valid)
		    next_receive[SEND_TO_HOST] = 1'b1;
		    //next_receive[IDLE] = 1'b1;
	     else
		    next_receive[READ_CHECK_BYTE] = 1'b1;
	  end 	     

	
`ifdef SIM
	if (state_receive == (1 << IDLE))
	  state_name_receive = "IDLE";
	else if (state_receive == (1 << RECEIVE_DATA))
	  state_name_receive = "RECEIVE_DATA";
	else if (state_receive == (1 << SEND_TO_HOST))
	  state_name_receive = "SEND_TO_HOST";
	else if (state_receive == (1 << RECEIVE_COMPLETE))
	  state_name_receive = "RECEIVE_COMPLETE";
	else if (state_receive == (1 << STORE_BYTE))
	  state_name_receive = "STORE_BYTE";
	else if (state_receive == (1 << SEND_CMD_BYTE))
	  state_name_receive = "SEND_CMD_BYTE";
	else if (state_receive == (1 << SEND_CMD_COMPLETE))
	  state_name_receive = "SEND_CMD_COMPLETE";
	else if (state_receive == (1 << CHECK_FOR_BYTE))
	  state_name_receive = "CHECK_FOR_BYTE";
	else if (state_receive == (1 << READ_CHECK_BYTE))
	  state_name_receive = "READ_CHECK_BYTE";
`endif
     end // always @ (...


   //--------------------------------------------------
   // SPI Write State Machine Transmit
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state_transmit <= 8'h00;
	     state_transmit[IDLE] <= 1'b1;
	  end
	else
	  state_transmit <= next_transmit;
     end

   always @ ( /*AUTOSENSE*/state_transmit or start_spi_write 
		 or USB_DATA or USB_RD_N or USB_RXF_N or USB_TXE_N
		 or spi_master_rdy_in or spi_delay_counter or spi_cs)
     begin
	next_transmit = 8'h00;

	if (state_transmit[IDLE])
	begin
        if(!start_spi_write)
			next_transmit[WRITE_SEND_CMD] = 1'b1;
		else
			next_transmit[IDLE] = 1'b1;
	end

	  if (state_transmit[WRITE_SEND_CMD])
	  begin
	     if (spi_master_rdy_in)
	       next_transmit[WRITE_CMD_COMPLETE] = 1'b1;
	     else 
	       next_transmit[WRITE_SEND_CMD] = 1'b1;
	  end

	  if (state_transmit[WRITE_CMD_COMPLETE])
	  begin
	     if (!spi_master_rdy_in)
	       next_transmit[WRITE_REGISTER] = 1'b1;
	     else 
	       next_transmit[WRITE_CMD_COMPLETE] = 1'b1;
	  end


	  if (state_transmit[WRITE_REGISTER])
	  begin
	     if (!USB_TXE_N)
	       next_transmit[LOAD_SHIFT_REG] = 1'b1;
	     else 
	       next_transmit[WRITE_REGISTER] = 1'b1;
	  end

	if (state_transmit[LOAD_SHIFT_REG])
	       next_transmit[DELAY_WRITE_DATA_BYTE] = 1'b1;

	if (state_transmit[DELAY_WRITE_DATA_BYTE])
	  begin
	       next_transmit[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state_transmit[WRITE_DATA_BYTE])
	  begin
        if(spi_delay_counter >= SPI_WRITE_DELAY_COUNT)
	       next_transmit[WRITE_DATA_COMPLETE] = 1'b1;
	    else
	       next_transmit[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state_transmit[WRITE_DATA_COMPLETE])
	  begin
		if (spi_cs)
	       next_transmit[IDLE] = 1'b1;
	     else
	       next_transmit[WRITE_DATA_COMPLETE] = 1'b1;
	  end

	
`ifdef SIM
	if (state_transmit == (1 << IDLE))
	  state_name_transmit = "IDLE";
	else if (state_transmit == (1 << WRITE_SEND_CMD))
	  state_name_transmit = "WRITE_SEND_CMD";
	else if (state_transmit == (1 << WRITE_CMD_COMPLETE))
	  state_name_transmit = "WRITE_CMD_COMPLETE";
	else if (state_transmit == (1 << WRITE_REGISTER))
	  state_name_transmit = "WRITE_REGISTER";
	else if (state_transmit == (1 << LOAD_SHIFT_REG))
	  state_name_transmit = "LOAD_SHIFT_REG";
	else if (state_transmit == (1 << DELAY_WRITE_DATA_BYTE))
	  state_name_transmit = "DELAY_WRITE_DATA_BYTE";
	else if (state_transmit == (1 << WRITE_DATA_BYTE))
	  state_name_transmit = "WRITE_DATA_BYTE";
	else if (state_transmit == (1 << WRITE_DATA_COMPLETE))
	  state_name_transmit = "WRITE_DATA_COMPLETE";
`endif
     end // always @ (...
	 

   //----------------------------------------------
   // Instantiate the SPI Top Level Leaf FSM Modules
   //----------------------------------------------
  spi_iface_master             SPI_IFACE_INST
  (
   .CLK                        (CLK),
   .RST_N                      (RST_N),
   .CLK_100                    (),
   .RST_N_100                  (),
   
   .SPI_DATA_IN                (miosio_miso),//(spi_miso),
   .SPI_CLK                    (spi_clk),
   .SPI_DATA_OUT               (spi_mosi),
   .SPI_CS_N                   (spi_cs),

   .DIN                        (spi_master_din), //SPI Master Write
   .WE_IN                      (spi_master_we_in),
   .RDY_IN                     (spi_master_rdy_in),

   .DOUT                      (spi_master_dout), //SPI Master Read
   .WE_OUT_N                  (spi_master_we_out_n),
   .RDY_OUT                   (spi_master_rdy_out)
 
   );	

endmodule // uart_slave

