//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2025
//#
//#
//# File Name:  rgb_led_control.v
//# Author:     Earth People Technology
//# Date:       Jan 16, 2025
//# Revision:   A
//#
//# Development: RGB LED Control
//# Application: Altera Cyclone 10 FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			1/16/25 	1			Created
//#
//#									
//#
//#######################################################################

module rgb_led_control(
	input  wire                 CLK_50MHZ,
	input  wire                 RST,

	output  reg                 RGB_LED_BLUE,
	output  reg                 RGB_LED_RED,
	output  reg                 RGB_LED_GREEN,

	input  wire [7:0]           LED_RGB_BYTE,
	input  wire                 LED_RGB_ENABLE,

	input  wire [7:0]           LED_DISPLAY_CONTROL


		);

//=======================================================================
// Parameter Declarations
//=======================================================================
   //LED State Machine Parameters
   parameter                   IDLE = 0,
	                           SELECT_MODE = 1,
                               LOAD_LEDS = 2,
                               LOAD_REGISTER = 3,
                               WAIT_FOR_TIMER = 4;
							   
   //Control Register selections for the LED Blinky
   parameter                   BLINK            = 4'ha;
   parameter                   SHIFT_LEFT       = 4'hb;
   parameter                   SHIFT_RIGHT      = 4'hc;
   parameter                   STATIC_VALUE     = 4'hd;
   parameter                   START_DISPLAY    = 4'he;
   parameter                   STOP_DISPLAY     = 4'hf;
   
   //RGB LED Turn On/Off 
   parameter                   ASSERT_RGB_LED_BLUE  = 1,
                               ASSERT_RGB_LED_RED   = 2,
                               ASSERT_RGB_LED_GREEN = 3,
                               DEASSERT_RGB_LED     = 4;
							   
   //Timer Value Low Limit
`ifdef SIM
   parameter                   TIMER_LOW_LIMIT       = 28'h1f;
   parameter                   HIGH_Z_TIMER_VALUE    = 28'h8f;
`else
   parameter                   TIMER_LOW_LIMIT       = 28'h2faf080;
   parameter                   HIGH_Z_TIMER_VALUE    = 28'ha1220; //10 ms
   //parameter                   HIGH_Z_TIMER_VALUE    = 28'h325aa0; //50 ms
`endif

`ifdef SIM
   	reg [8*30:1] state_name;
`endif

//=======================================================================
// Register Declarations
//=======================================================================
	//Finite State Machine control registers
	reg  [4:0]                 state, next;

	//LED BLinky Signals
	reg  [27:0]                timer_value;	
	reg                        start_blinky;
	reg  [7:0]                 static_load_value;	
	reg  [7:0]                 shift_count_value;
	wire [7:0]                 control_register;
	reg  [27:0]                led_delay_counter;
	reg                        led_display_register_1;
	reg                        led_display_register_2;
	reg                        led_display_register_3;
	reg                        shift_count_value_1;
	reg                        shift_count_value_2;
	reg                        shift_count_value_3;
	

//=======================================================================
// Simulation Directives
//=======================================================================

//=======================================================================
// Assignments	
//=======================================================================
   assign control_register = LED_DISPLAY_CONTROL;
   //-----------------------------------------------
   // Set the LED outputs 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	begin
       RGB_LED_BLUE <= 1'bz;
       RGB_LED_RED <= 1'bz;
       RGB_LED_GREEN <= 1'bz;
    end
	else 
	begin
	  if(state[LOAD_LEDS])
	  begin
     //****************************************************
	 // RGB BLUE LED 
     //****************************************************

	    //LED Display 1 Segment A
		if(led_display_register_1)
			RGB_LED_BLUE <= 1'b0;
		else 
			RGB_LED_BLUE <= 1'bz;
		
     //****************************************************
	 // RGB RED LED 
     //****************************************************

	 //LED Display 2 Segment A		
		if(led_display_register_2)
			RGB_LED_RED <= 1'b0;
		else 
			RGB_LED_RED <= 1'bz;
						
     //****************************************************
	 // LED Display 3
     //****************************************************

        //LED Display 3 Segment A		
		if(led_display_register_3)
			RGB_LED_GREEN <= 1'b0;
		else 
			RGB_LED_GREEN <= 1'bz;
			
		end
     end
  end
   //-----------------------------------------------
   // LED Blinky Load Timer Value
   //-----------------------------------------------

  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	     timer_value <= TIMER_LOW_LIMIT;
	else 
	begin
	   timer_value <= TIMER_LOW_LIMIT;
/*	     if(trigger_in_byte[4])
	         timer_value[15:8] <= transfer_in_byte_userled;
	     else if(trigger_in_byte[5])
	         timer_value[23:16] <= transfer_in_byte_userled;
*/	end 
  end
  
   //-----------------------------------------------
   // LED Blinky start 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
         start_blinky <= 1'b0;
	else 
	begin
	   if((control_register[3:0] == START_DISPLAY) | LED_RGB_ENABLE)
         start_blinky <= 1'b1;
	   else if(control_register[3:0] == STOP_DISPLAY) 
	     start_blinky <= 1'b0;
 	end 
  end
 
   //-----------------------------------------------
   // LED Blinky Delay Timer Counter 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	     led_delay_counter <= TIMER_LOW_LIMIT;
	else 
	begin
	     if(state[SELECT_MODE])
		 led_delay_counter <= timer_value;
		 else if(state[WAIT_FOR_TIMER])
	         led_delay_counter <= led_delay_counter - 1'd1;
	end 
  end

   //-----------------------------------------------
   // Select the LED Function from the control
   // register bits 7 to 4   
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
    begin
       led_display_register_1  <= 1'b0;
       led_display_register_2  <= 1'b0;
       led_display_register_3  <= 1'b0;
    end
	else 
	begin
	    if(LED_RGB_ENABLE)
		begin
			case(LED_RGB_BYTE) 
		       ASSERT_RGB_LED_BLUE:
		          led_display_register_1 <= 1'b1;
		       ASSERT_RGB_LED_RED:
		          led_display_register_2 <= 1'b1;
		       ASSERT_RGB_LED_GREEN:
		          led_display_register_3 <= 1'b1;
		       DEASSERT_RGB_LED:
			   begin
                  led_display_register_1 <= 1'b0;
                  led_display_register_2 <= 1'b0;
                  led_display_register_3 <= 1'b0;
			   end
		       default:
			   begin
                  led_display_register_1 <= 1'b0;
                  led_display_register_2 <= 1'b0;
                  led_display_register_3 <= 1'b0;
			   end
		   endcase 
        end
		else if (state[SELECT_MODE])
	    begin
	         //case(control_register[3:0])
		     //  SHIFT_LEFT:
			 //  begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
/*			   end
		       SHIFT_RIGHT:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
			   end
		       COUNT_UP:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
			   end
		       COUNT_DOWN:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		       STATIC_VALUE:
			   begin
		          led_display_register_1  = shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		       default:
			   begin
		          led_display_register_1  = shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		      endcase 
*/        end	
    end
  end
  
   //-----------------------------------------------
   // Increment the Shift/Count value from the control
   // register bits 7 to 4   
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	begin
	     shift_count_value_1 <= 1'b0;
	     shift_count_value_2 <= 1'b0;
	     shift_count_value_3 <= 1'b0;
	end
	else 
	begin
	    if(LED_RGB_ENABLE)
		begin
			case(LED_RGB_BYTE)
		       ASSERT_RGB_LED_BLUE:
		          shift_count_value_1 <= 1'b1;
		       ASSERT_RGB_LED_RED:
		          shift_count_value_2 <= 1'b1;
		       ASSERT_RGB_LED_GREEN:
		          shift_count_value_3 <= 1'b1;
		       DEASSERT_RGB_LED:
			   begin
                  shift_count_value_1 <= 1'b0;
                  shift_count_value_2 <= 1'b0;
                  shift_count_value_3 <= 1'b0;
			   end
		       default:
			   begin
                  shift_count_value_1 <= 1'b0;
                  shift_count_value_2 <= 1'b0;
                  shift_count_value_3 <= 1'b0;
			   end
		   endcase 
        end
        else if (state[LOAD_REGISTER])
	    begin
	         //case(control_register[3:0])
		     //SHIFT_LEFT:
			 //begin
		        shift_count_value_1 <= led_display_register_3;
		        shift_count_value_3 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_1;
/*		     end
		     SHIFT_RIGHT:
			 begin
		        shift_count_value_3 <= led_display_register_1;
		        shift_count_value_2 <= led_display_register_3;
		        shift_count_value_1 <= led_display_register_2;
		     end
		     COUNT_UP:
			 begin
		        shift_count_value_1 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_3;
		        shift_count_value_3 <= led_display_register_1;
		     end
		     COUNT_DOWN:
			 begin
		        shift_count_value_3 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_1;
		        shift_count_value_1 <= led_display_register_3;
		     end
			 STATIC_VALUE:
			 begin
		        shift_count_value_3 <= led_display_register_3;
		        shift_count_value_2 <= led_display_register_2;
		        shift_count_value_1 <= led_display_register_1;
		     end
		     default:
		     //SHIFT_LEFT:
			 begin
		        shift_count_value_3 <= shift_count_value_3;
		        shift_count_value_2 <= shift_count_value_2;
		        shift_count_value_1 <= shift_count_value_1;
		     end
		    endcase   
*/         end
	end 
  end
  
     //-----------------------------------------------
   // State Machine: Send Pattern to LEDs 
   //-----------------------------------------------
   // Next State Logic
   always @(posedge CLK_50MHZ or negedge RST)
     begin
	if (!RST)
	  begin
	     state <= 5'h00;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

     // State Definitions
   always @ ( state or  start_blinky or
		      led_display_register_1 or led_delay_counter or 
			  led_display_register_2 or timer_value or shift_count_value or
			  led_display_register_3) 
     begin
	     next = 5'h00;

	if (state[IDLE])
	  begin
	     if (start_blinky)
	       next[SELECT_MODE] = 1'b1;
	     else
	       next[IDLE] = 1'b1;
	  end
	
	if (state[SELECT_MODE])
	begin
       next[LOAD_REGISTER] = 1'b1;
	end

	if (state[LOAD_REGISTER])
	begin
        next[LOAD_LEDS] = 1'b1;
    end

	if (state[LOAD_LEDS])
	begin
        next[WAIT_FOR_TIMER] = 1'b1;
    end
			
	if (state[WAIT_FOR_TIMER])
		//if (led_delay_counter <= timer_value)
		if (led_delay_counter > 0)
			next[WAIT_FOR_TIMER] = 1'b1;
		else 
			next[SELECT_MODE] = 1'b1;

			
`ifdef SIM
 	   if ( state == ( 1 << IDLE ))
		   state_name = "IDLE";
	   else if ( state == ( 1 << SELECT_MODE ))
		   state_name = "SELECT_MODE";
	   else if ( state == ( 1 << LOAD_REGISTER ))
		   state_name = "LOAD_REGISTER";
	   else if ( state == ( 1 << LOAD_LEDS ))
		   state_name = "LOAD_LEDS";
	   else if ( state == ( 1 << WAIT_FOR_TIMER ))
		   state_name = "WAIT_FOR_TIMER";
`endif


	end//end of state machine





endmodule

