//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2024
//#
//#
//# File Name:  led_update.v
//# Author:     Earth People Technology
//# Date:       March 20, 2019
//# Revision:   A
//#
//# Development: EPT I2C Slave Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			10/5/24 	1			Created
//#
//#									
//#
//#######################################################################

module seven_segment_led_display(
	input  wire                 CLK_50MHZ,
	input  wire                 RST,

	input  wire [7:0]           LED_DISPLAY_BYTE_1,
	input  wire                 LED_DISPLAY_ENABLE_1,

	input  wire [7:0]           LED_DISPLAY_BYTE_2,
	input  wire                 LED_DISPLAY_ENABLE_2,

	input  wire [7:0]           LED_DISPLAY_BYTE_3,
	input  wire                 LED_DISPLAY_ENABLE_3,

	output reg  [7:0]           SEVEN_SEGMENT_LED_1,
	output reg  [7:0]           SEVEN_SEGMENT_LED_2,
	output reg  [7:0]           SEVEN_SEGMENT_LED_3,
	
	input  wire [7:0]           LED_DISPLAY_CONTROL


		);

//=======================================================================
// Parameter Declarations
// SEVEN_SEGMENT_LED_x{7] = Segment A
// SEVEN_SEGMENT_LED_x{6] = Segment B
// SEVEN_SEGMENT_LED_x{5] = Segment C
// SEVEN_SEGMENT_LED_x{4] = Segment D
// SEVEN_SEGMENT_LED_x{3] = Segment E
// SEVEN_SEGMENT_LED_x{2] = Segment F
// SEVEN_SEGMENT_LED_x{1] = Segment G
// SEVEN_SEGMENT_LED_x{0] = Segment DP
//=======================================================================
   //LED State Machine Parameters
   parameter                   IDLE = 0,
	                           SELECT_MODE = 1,
                               LOAD_LEDS = 2,
                               LOAD_REGISTER = 3,
                               WAIT_FOR_TIMER = 4;
							   
   //Control Register selections for the LED Blinky
   parameter                   BLINK            = 4'h1;
   parameter                   SHIFT_LEFT       = 4'h2;
   parameter                   SHIFT_RIGHT      = 4'h3;
   parameter                   COUNT_UP         = 4'h4;
   parameter                   COUNT_DOWN       = 4'h5;
   parameter                   STATIC_VALUE     = 4'h6;
   parameter                   START_DISPLAY    = 4'h7;
   parameter                   STOP_DISPLAY     = 4'h8;

   //Timer Value Low Limit
`ifdef SIM
   parameter                   TIMER_LOW_LIMIT       = 28'h1f;
   parameter                   HIGH_Z_TIMER_VALUE    = 28'h8f;
`else
   parameter                   TIMER_LOW_LIMIT       = 28'h2faf080;
   parameter                   HIGH_Z_TIMER_VALUE    = 28'ha1220; //10 ms
   //parameter                   HIGH_Z_TIMER_VALUE    = 28'h325aa0; //50 ms
`endif

`ifdef SIM
   	reg [8*30:1] state_name;
`endif

//=======================================================================
// Register Declarations
//=======================================================================
	//Finite State Machine control registers
	reg  [4:0]                 state, next;

	//LED BLinky Signals
	reg  [27:0]                timer_value;	
	reg                        start_blinky;
	reg  [7:0]                 static_load_value;	
	reg  [7:0]                 shift_count_value;
	reg  [7:0]                 led_reg;
	wire [7:0]                 control_register;
	reg  [27:0]                led_delay_counter;
	reg  [7:0]                 led_display_register_1;
	reg  [7:0]                 led_display_register_2;
	reg  [7:0]                 led_display_register_3;
	reg  [7:0]                 shift_count_value_1;
	reg  [7:0]                 shift_count_value_2;
	reg  [7:0]                 shift_count_value_3;
	

//=======================================================================
// Simulation Directives
//=======================================================================

//=======================================================================
// Assignments	
//=======================================================================
   assign control_register = LED_DISPLAY_CONTROL;
   //-----------------------------------------------
   // Set the LED outputs 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	begin
	     SEVEN_SEGMENT_LED_1 = 8'h61;
	     SEVEN_SEGMENT_LED_2 = 8'h31;
	     SEVEN_SEGMENT_LED_3 = 8'h11;
    end
	else 
	begin
	  if(state[LOAD_LEDS])
	  begin
     //****************************************************
	 // LED Display 1
     //****************************************************

	    //LED Display 1 Segment A
		if ( led_display_register_1[7] )
			SEVEN_SEGMENT_LED_1[0] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[0] <= 1'bz;
		
        //LED Display 1 Segment B		
		if ( led_display_register_1[6] )
			SEVEN_SEGMENT_LED_1[1] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[1] <= 1'bz;
			
        //LED Display 1 Segment C		
 		if ( led_display_register_1[5] )
			SEVEN_SEGMENT_LED_1[2] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[2] <= 1'bz;
			
        //LED Display 1 Segment D		
		if ( led_display_register_1[4] )
			SEVEN_SEGMENT_LED_1[3] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[3] <= 1'bz;
			
        //LED Display 1 Segment E		
		if ( led_display_register_1[3] )
			SEVEN_SEGMENT_LED_1[4] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[4] <= 1'bz;
			
        //LED Display 1 Segment F		
		if ( led_display_register_1[2] )
			SEVEN_SEGMENT_LED_1[5] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[5] <= 1'bz;
			
        //LED Display 1 Segment G		
 		if ( led_display_register_1[1] )
			SEVEN_SEGMENT_LED_1[6] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[6] <= 1'bz;
			
        //LED Display 1 Segment DP		
		if ( led_display_register_1[0] )
			SEVEN_SEGMENT_LED_1[7] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_1[7] <= 1'bz;
     //****************************************************
	 // LED Display 2
     //****************************************************

	 //LED Display 2 Segment A		
		if (led_display_register_2[7] )
			SEVEN_SEGMENT_LED_2[0] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[0] <= 1'bz;
			
        //LED Display 2 Segment B		
		if (led_display_register_2[6] )
			SEVEN_SEGMENT_LED_2[1] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[1] <= 1'bz;
			
        //LED Display 2 Segment C		
 		if (led_display_register_2[5] )
			SEVEN_SEGMENT_LED_2[2] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[2] <= 1'bz;
			
        //LED Display 2 Segment D		
		if (led_display_register_2[4] )
			SEVEN_SEGMENT_LED_2[3] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[3] <= 1'bz;
			
        //LED Display 2 Segment E		
		if (led_display_register_2[3] )
			SEVEN_SEGMENT_LED_2[4] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[4] <= 1'bz;
			
        //LED Display 2 Segment F		
		if (led_display_register_2[2])
			SEVEN_SEGMENT_LED_2[5] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[5] <= 1'bz;
			
        //LED Display 2 Segment G		
 		if (led_display_register_2[1] )
			SEVEN_SEGMENT_LED_2[6] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[6] <= 1'bz;
			
        //LED Display 2 Segment DP		
		if (led_display_register_2[0] )
			SEVEN_SEGMENT_LED_2[7] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_2[7] <= 1'bz;
			
     //****************************************************
	 // LED Display 3
     //****************************************************

        //LED Display 3 Segment A		
		if (led_display_register_3[7] )
			SEVEN_SEGMENT_LED_3[0] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[0] <= 1'bz;
			
        //LED Display 3 Segment B		
		if (led_display_register_3[6] )
			SEVEN_SEGMENT_LED_3[1] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[1] <= 1'bz;

        //LED Display 3 Segment C		
		if (led_display_register_3[5] )
			SEVEN_SEGMENT_LED_3[2] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[2] <= 1'bz;
			
        //LED Display 3 Segment D		
 		if (led_display_register_3[4] )
			SEVEN_SEGMENT_LED_3[3] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[3] <= 1'bz;
			
        //LED Display 3 Segment E		
		if (led_display_register_3[3] )
			SEVEN_SEGMENT_LED_3[4] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[4] <= 1'bz;
			
        //LED Display 3 Segment F		
		if (led_display_register_3[2] )
			SEVEN_SEGMENT_LED_3[5] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[5] <= 1'bz;
			
        //LED Display 3 Segment G		
		if (led_display_register_3[1] )
			SEVEN_SEGMENT_LED_3[6] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[6] <= 1'bz;
			
        //LED Display 3 Segment DP		
 		if (led_display_register_3[0] )
			SEVEN_SEGMENT_LED_3[7] <= 1'b0;
		else 
			SEVEN_SEGMENT_LED_3[7] <= 1'bz;
	   end	
	end
  end

   //-----------------------------------------------
   // LED Blinky Load Timer Value
   //-----------------------------------------------

  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	     timer_value <= TIMER_LOW_LIMIT;
	else 
	begin
	   timer_value <= TIMER_LOW_LIMIT;
/*	     if(trigger_in_byte[4])
	         timer_value[15:8] <= transfer_in_byte_userled;
	     else if(trigger_in_byte[5])
	         timer_value[23:16] <= transfer_in_byte_userled;
*/	end 
  end
  
   //-----------------------------------------------
   // LED Blinky start 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
         start_blinky <= 1'b0;
	else 
	begin
	   if(control_register[3:0] == START_DISPLAY)
         start_blinky <= 1'b1;
	   else if(control_register[3:0] == STOP_DISPLAY) 
	     start_blinky <= 1'b0;
 	end 
  end
 
   //-----------------------------------------------
   // LED Blinky Delay Timer Counter 
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	     led_delay_counter <= TIMER_LOW_LIMIT;
	else 
	begin
	     if(state[SELECT_MODE])
		 led_delay_counter <= timer_value;
		 else if(state[WAIT_FOR_TIMER])
	         led_delay_counter <= led_delay_counter - 1'd1;
	end 
  end

   //-----------------------------------------------
   // Select the LED Function from the control
   // register bits 7 to 4   
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
    begin
       led_display_register_1  <= 8'h73;
       led_display_register_2  <= 8'h77;
       led_display_register_3  <= 8'h79;
    end
	else 
	begin
	    if(LED_DISPLAY_ENABLE_1)
		begin
			 led_display_register_1 <= {LED_DISPLAY_BYTE_1[0],LED_DISPLAY_BYTE_1[1],LED_DISPLAY_BYTE_1[2],LED_DISPLAY_BYTE_1[3],LED_DISPLAY_BYTE_1[4],LED_DISPLAY_BYTE_1[5],LED_DISPLAY_BYTE_1[6],LED_DISPLAY_BYTE_1[7]};
		end
	    else if(LED_DISPLAY_ENABLE_2)
		begin
			 led_display_register_2 <= {LED_DISPLAY_BYTE_2[0],LED_DISPLAY_BYTE_2[1],LED_DISPLAY_BYTE_2[2],LED_DISPLAY_BYTE_1[3],LED_DISPLAY_BYTE_2[4],LED_DISPLAY_BYTE_2[5],LED_DISPLAY_BYTE_2[6],LED_DISPLAY_BYTE_2[7]};
		end
	    else if(LED_DISPLAY_ENABLE_3)
		begin
			 led_display_register_3 <= {LED_DISPLAY_BYTE_3[0],LED_DISPLAY_BYTE_3[1],LED_DISPLAY_BYTE_3[2],LED_DISPLAY_BYTE_3[3],LED_DISPLAY_BYTE_3[4],LED_DISPLAY_BYTE_3[5],LED_DISPLAY_BYTE_3[6],LED_DISPLAY_BYTE_3[7]};
		end
		else if (state[SELECT_MODE])
	    begin
	         //case(control_register[3:0])
		     //  SHIFT_LEFT:
			 //  begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
/*			   end
		       SHIFT_RIGHT:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
			   end
		       COUNT_UP:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  <= shift_count_value_2;
		          led_display_register_3  <= shift_count_value_3;
			   end
		       COUNT_DOWN:
			   begin
		          led_display_register_1  <= shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		       STATIC_VALUE:
			   begin
		          led_display_register_1  = shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		       default:
			   begin
		          led_display_register_1  = shift_count_value_1;
		          led_display_register_2  = shift_count_value_2;
		          led_display_register_3  = shift_count_value_3;
			   end
		      endcase 
*/        end	
    end
  end
  
   //-----------------------------------------------
   // Increment the Shift/Count value from the control
   // register bits 7 to 4   
   //-----------------------------------------------
  always @(posedge CLK_50MHZ or negedge RST)
  begin
    if(!RST)
	begin
	     shift_count_value_1 <= 8'h73;
	     shift_count_value_2 <= 8'h77;
	     shift_count_value_3 <= 8'h79;
	end
	else 
	begin
	    if(LED_DISPLAY_ENABLE_1)
		begin
			 shift_count_value_1 <= {LED_DISPLAY_BYTE_1[0],LED_DISPLAY_BYTE_1[1],LED_DISPLAY_BYTE_1[2],LED_DISPLAY_BYTE_1[3],LED_DISPLAY_BYTE_1[4],LED_DISPLAY_BYTE_1[5],LED_DISPLAY_BYTE_1[6],LED_DISPLAY_BYTE_1[7]};
		end
	    else if(LED_DISPLAY_ENABLE_2)
		begin
			 shift_count_value_2 <= {LED_DISPLAY_BYTE_2[0],LED_DISPLAY_BYTE_2[1],LED_DISPLAY_BYTE_2[2],LED_DISPLAY_BYTE_1[3],LED_DISPLAY_BYTE_2[4],LED_DISPLAY_BYTE_2[5],LED_DISPLAY_BYTE_2[6],LED_DISPLAY_BYTE_2[7]};
		end
	    else if(LED_DISPLAY_ENABLE_3)
		begin
			 shift_count_value_3 <= {LED_DISPLAY_BYTE_3[0],LED_DISPLAY_BYTE_3[1],LED_DISPLAY_BYTE_3[2],LED_DISPLAY_BYTE_3[3],LED_DISPLAY_BYTE_3[4],LED_DISPLAY_BYTE_3[5],LED_DISPLAY_BYTE_3[6],LED_DISPLAY_BYTE_3[7]};
		end
		else if (state[LOAD_REGISTER])
	    begin
	         //case(control_register[3:0])
		     //SHIFT_LEFT:
			 //begin
		        shift_count_value_1 <= led_display_register_3;
		        shift_count_value_3 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_1;
/*		     end
		     SHIFT_RIGHT:
			 begin
		        shift_count_value_3 <= led_display_register_1;
		        shift_count_value_2 <= led_display_register_3;
		        shift_count_value_1 <= led_display_register_2;
		     end
		     COUNT_UP:
			 begin
		        shift_count_value_1 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_3;
		        shift_count_value_3 <= led_display_register_1;
		     end
		     COUNT_DOWN:
			 begin
		        shift_count_value_3 <= led_display_register_2;
		        shift_count_value_2 <= led_display_register_1;
		        shift_count_value_1 <= led_display_register_3;
		     end
			 STATIC_VALUE:
			 begin
		        shift_count_value_3 <= led_display_register_3;
		        shift_count_value_2 <= led_display_register_2;
		        shift_count_value_1 <= led_display_register_1;
		     end
		     default:
		     //SHIFT_LEFT:
			 begin
		        shift_count_value_3 <= shift_count_value_3;
		        shift_count_value_2 <= shift_count_value_2;
		        shift_count_value_1 <= shift_count_value_1;
		     end
		    endcase   
*/         end
	end 
  end
  
     //-----------------------------------------------
   // State Machine: Send Pattern to LEDs 
   //-----------------------------------------------
   // Next State Logic
   always @(posedge CLK_50MHZ or negedge RST)
     begin
	if (!RST)
	  begin
	     state <= 5'h00;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

     // State Definitions
   always @ ( state or  start_blinky or
		      led_reg or led_delay_counter or 
			  timer_value or shift_count_value) 
     begin
	     next = 5'h00;

	if (state[IDLE])
	  begin
	     if (start_blinky)
	       next[SELECT_MODE] = 1'b1;
	     else
	       next[IDLE] = 1'b1;
	  end
	
	if (state[SELECT_MODE])
	begin
       next[LOAD_REGISTER] = 1'b1;
	end

	if (state[LOAD_REGISTER])
	begin
        next[LOAD_LEDS] = 1'b1;
    end

	if (state[LOAD_LEDS])
	begin
        next[WAIT_FOR_TIMER] = 1'b1;
    end
			
	if (state[WAIT_FOR_TIMER])
		//if (led_delay_counter <= timer_value)
		if (led_delay_counter > 0)
			next[WAIT_FOR_TIMER] = 1'b1;
		else 
			next[SELECT_MODE] = 1'b1;

			
`ifdef SIM
 	   if ( state == ( 1 << IDLE ))
		   state_name = "IDLE";
	   else if ( state == ( 1 << SELECT_MODE ))
		   state_name = "SELECT_MODE";
	   else if ( state == ( 1 << LOAD_REGISTER ))
		   state_name = "LOAD_REGISTER";
	   else if ( state == ( 1 << LOAD_LEDS ))
		   state_name = "LOAD_LEDS";
	   else if ( state == ( 1 << WAIT_FOR_TIMER ))
		   state_name = "WAIT_FOR_TIMER";
`endif


	end//end of state machine





endmodule

