//================================================================
//
// 	Copyright 	Earth People Technology Inc. 2021
//
//
//	File Name:	ams1115_model.v
//	Author:	    R. Jolly
//	Date:		May 25, 2021
//	Revision:	A
//
//
//  Development: EPT I2C Slave Project 
//  Application: Altera Cyclone IV FPGA
//  Description: This file contains verilog code which will allow access
//               to Active Transfer Library. 
//				
//
//
//
//	Revision History:
//
//	DATE		VERSION		DETAILS
//	5/25/21		A		Created by R Jolly
//
//
//
//
//
//  Register Mapping:
//           STATUS_REGISTER[0] = Read a Receive word from the I2C Unit
//           STATUS_REGISTER[1] = Write a Transmit word to the I2C Unit
//           STATUS_REGISTER[2] = Latch the Address into the I2C Unit
//           STATUS_REGISTER[3] = Request Recieve words from the I2C Unit
//           STATUS_REGISTER[4] = 
//           STATUS_REGISTER[5] = 
//           STATUS_REGISTER[6] = 
//           STATUS_REGISTER[7] = 
//
//
//
//
//
//==================================================================

module ams1115_model
(
	input wire                  CLK,
	input wire                  RESET,

	input wire   [7:0]          ADDRESS,

	inout wire                  DATA,
	input reg                   CLOCK,

	output reg   [7:0]          CONTROL_REGISTER,
	input  wire  [7:0]          STATUS_REGISTER,
	output wire  [7:0]          RECEIVE_REGISTER,
	input  wire  [7:0]          SEND_REGISTER

		);

//==================================================================
// Parameter Declarations
//==================================================================
    parameter   IDLE                        = 0,
                ASSERT_ENA_IN               = 1,
                READ_DATA                   = 2,
                DE_ASSERT_ENA_IN            = 3,
                ASSERT_RDY_IN               = 4,
                TRANSMIT_DATA               = 5,
                DE_ASSERT_RDY_IN            = 6;

//==================================================================
// IO Port Declarations
//==================================================================

//==================================================================
// Register Declarations
//==================================================================
  
	// State Machine Registers
	reg  [7:0]					state;
	reg  [7:0]					next;

	// Transmit Registers for Serial I2C Comms
	wire  [7:0]                 transmit_data_in;
	reg                         RDY_OUT;
	wire  [7:0]                 word_count;
	
	// Receive Registers for Serial I2C Comms
	wire  [7:0]                 receive_data_out;
	reg                         ENA_OUT;
	wire                        RDY_IN;

	// de_assert_ena_in_count register
	reg  [7:0]                  de_assert_ena_in_count;

	// read_data_count register
	reg  [7:0]                  read_data_count;

	// CONTROL Register
	reg                         control_update_reg;
	reg                         status_update_reg;

	// ENA_OUT register, this register acknowledges the RDY_IN from
	// serial_i2c_comms.v and causes the function to update its fsm
//	reg                         ENA_OUT;
	reg                         ena_out_reg;

	// receive_data_ready  registers, used to initiate the local fsm
	reg                         receive_data_ready;
	reg                         receive_data_ready_reg;

	// Transmit data counter
	reg	[7:0]					transmit_data_count;

	// De-assert the RDY_OUT counter
	reg	[7:0]					de_assert_rdy_in_count;

	// Ready Out registers
	reg                         rdy_out_reg;

	// Transmit data ready registers
	reg                         transmit_data_ready;
	reg                         transmit_data_ready_reg;

	// Enable out register from serial_i2c_comms
	wire                        ENA_IN;
	
	//Address Registers
	reg  [7:0]                  i2c_address;
	reg                         store_address_reg;
	
	//Receive Request registers
	reg  [7:0]                  receive_req_word_count;
	reg                         receive_req_word_count_reg;
	


//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

	assign	transmit_data_in = STATUS_REGISTER[2] ? 8'h00 : SEND_REGISTER;
	assign	word_count = receive_req_word_count;

	assign	RECEIVE_REGISTER = state[READ_DATA] ? receive_data_out : RECEIVE_REGISTER;
	

//==================================================================
// Transmit Cycle: Request a receive cycle from the I2C Unit
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			receive_req_word_count <= 0;
			receive_req_word_count_reg <= 1'b0;
		end
		else 
		begin
			if ( STATUS_REGISTER[3] & !receive_req_word_count_reg )
			begin
				receive_req_word_count <= SEND_REGISTER;
				receive_req_word_count_reg <= 1'b1;
			end
			else if ( !STATUS_REGISTER[3] & receive_req_word_count_reg )
			begin
				receive_req_word_count <= 0;
				receive_req_word_count_reg <= 1'b0;
			end
			else if(state[IDLE])
			begin
				receive_req_word_count_reg <= 1'b0;
			end
		end		
	end // always @ (posedge CLK or negedge RESET)


//==================================================================
// Transmit Cycle: Store the I2C Address when STATUS_REGISTER[2]
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			i2c_address <= 0;
			store_address_reg <= 1'b0;
		end
		else 
		begin
			if ( STATUS_REGISTER[2] & !store_address_reg )
			begin
				i2c_address <= SEND_REGISTER;
				store_address_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				store_address_reg <= 1'b0;
			end
		end		
	end // always @ (posedge CLK or negedge RESET)

//=================================================================================
//  Transmit Cycle: de_assert_rdy_in_count 
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			de_assert_rdy_in_count <= 8'h0;
		end
		else
		begin
        		if ( state[DE_ASSERT_RDY_IN] )
			begin
             		de_assert_rdy_in_count <= de_assert_rdy_in_count + 1;
			end
			else if ( state[IDLE] )
			begin
				de_assert_rdy_in_count <= 8'h0;
			end
				
    		end
  end

//=================================================================================
//  Transmit Cycle: transmit_data_count 
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
      		transmit_data_count <= 8'h0;
    		end
  		else
    		begin
        		if ( state[TRANSMIT_DATA] )
          		begin
              		transmit_data_count <= transmit_data_count  + 1;
          		end
                else if ( state[IDLE] )
            		transmit_data_count <= 8'h0;
    		end
	end

//=================================================================================
//	Transmit Cycle: RDY_OUT
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			RDY_OUT <= 1'b0;
			rdy_out_reg <= 1'b0;
		end
		else
		begin
			if( state[ASSERT_RDY_IN] & !rdy_out_reg )
			begin
				RDY_OUT <= 1'b1;
				rdy_out_reg <= 1'b1;
			end
			else if ( state[DE_ASSERT_RDY_IN] & ENA_IN & rdy_out_reg )
			begin
				RDY_OUT <=1'b0;
				rdy_out_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				RDY_OUT <=1'b0;
				rdy_out_reg <= 1'b0;
			end
		end
	end


//==================================================================
// Transmit Cycle: transmit_data_ready will be equated with STATUS_REGISTER[1]
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			transmit_data_ready <= 0;
			transmit_data_ready_reg <= 0;
		end
		else 
		begin
			if ( STATUS_REGISTER[1] & !transmit_data_ready_reg )
			begin
				transmit_data_ready <= 1'b1;
				transmit_data_ready_reg <= 1'b1;
			end
			else if ( STATUS_REGISTER[1] & transmit_data_ready_reg )
			begin
				transmit_data_ready <= 1'b0;
				transmit_data_ready_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				transmit_data_ready <= 0;
				transmit_data_ready_reg <= 0;
			end
		end		
	end // always @ (posedge CLK or negedge RESET)
   
//=================================================================================
//  Read Cycle: de_assert_ena_in_count
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			de_assert_ena_in_count <= 8'h0;
		end
		else
		begin
            if ( state[DE_ASSERT_ENA_IN] )
			begin
                de_assert_ena_in_count <= de_assert_ena_in_count + 1;
			end
			else if ( state[IDLE] )
			begin
				de_assert_ena_in_count <= 8'h0;
			end
				
    		end
  end

//=================================================================================
//  Read Cycle: read_data_count 
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
      		read_data_count <= 8'h0;
    		end
  		else
    		begin
        		if ( state[READ_DATA] )
          		begin
              		read_data_count <= read_data_count + 1;
          		end
       		else if ( state[DE_ASSERT_ENA_IN] )
            		read_data_count <= 8'h0;
    		end
	end



//=================================================================================
//	Receive Cycle: CONTROL_REGISTER, this register informs the ARM9 when sensor
//	data has been updated.
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			CONTROL_REGISTER <= 8'b0;
			control_update_reg <= 1'b0;
			status_update_reg <= 1'b0;
		end
		else
		begin
			if( STATUS_REGISTER[0] == 1'b1 & !status_update_reg )
			begin
				CONTROL_REGISTER[2] <= 1'b0;
				status_update_reg <= 1'b1;
			end
			else if( STATUS_REGISTER[0] == 1'b1 & status_update_reg )
			begin
				status_update_reg <= 1'b1;
			end
			else if( STATUS_REGISTER[0] == 1'b0 & status_update_reg )
			begin
				status_update_reg <= 1'b0;
			end

			if ( state[READ_DATA] & !control_update_reg )
			begin
				CONTROL_REGISTER[2] <= CONTROL_REGISTER[2] || 1'b1;
				control_update_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				control_update_reg <= 1'b0;
			end
		end
	end


//=================================================================================
//	Receive Cycle: ENA_OUT
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			ENA_OUT <= 1'b0;
			ena_out_reg <= 1'b0;
		end
		else
		begin
			if( state[ASSERT_ENA_IN] & !ena_out_reg )
			begin
				ENA_OUT <= 1'b1;
				ena_out_reg <= 1'b1;
			end
			else if ( state[DE_ASSERT_ENA_IN] & ena_out_reg )
			begin
				ENA_OUT <=1'b0;
				ena_out_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				ENA_OUT <=1'b0;
				ena_out_reg <= 1'b0;
			end
		end
	end




//=================================================================================
//	Receive Cycle: Create the receive_data_ready for initiating the receive word 
//	from serial_i2c_comms.v
//=================================================================================
	always @( posedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_data_ready <= 1'b0;
			receive_data_ready_reg  <= 1'b0;
		end
		else
		begin
			if( RDY_IN & ( CONTROL_REGISTER[2] == 1'b0 ) & !receive_data_ready_reg )
			begin
				receive_data_ready <= 1'b1;
				receive_data_ready_reg <= 1'b1;
			end
			else if( RDY_IN & receive_data_ready_reg )
			begin
				receive_data_ready <= 1'b0;
				receive_data_ready_reg <= 1'b1;
			end
//			else if( !RDY_IN & receive_data_ready_reg )
//			begin
//				receive_data_ready <= 1'b0;
//				receive_data_ready_reg <= 1'b0;
//			end
			else if( state[IDLE] )
				receive_data_ready_reg <= 1'b0;

		end
	end

//=================================================================================
//	Serial I2C Communications Finite State Machine
//	parameter    IDLE                     = 0,
//               ASSERT_ENA_IN            = 1,
//               READ_DATA                = 2,
//               DE_ASSERT_ENA_IN         = 3,
//               ASSERT_RDY_IN            = 4,
//               TRANSMIT_DATA            = 5,
//               DE_ASSERT_RDY_IN         = 6;
//
//=================================================================================

	always @(  posedge CLK or negedge RESET  )
	begin
		if( !RESET )
		begin
			state <= 8'h00;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( state or receive_data_ready or RDY_IN or ENA_OUT or 
		read_data_count or de_assert_ena_in_count or
		transmit_data_count or de_assert_rdy_in_count or rdy_out_reg or
		transmit_data_ready)
	begin
		next = 8'h00;
		if ( state[IDLE] )
		begin
			if ( receive_data_ready )
				next[ASSERT_ENA_IN] = 1'b1;
			else if (transmit_data_ready )
				next[ASSERT_RDY_IN] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if ( state[ASSERT_ENA_IN] )
		begin
			if ( ENA_OUT )
				next[READ_DATA] = 1'b1;
			else
				next[ASSERT_ENA_IN] = 1'b1;
		end

		if ( state[READ_DATA] )
		begin
			if ( read_data_count >= 8'h8 )
				next[DE_ASSERT_ENA_IN] = 1'b1;
			else
				next[READ_DATA] = 1'b1;
		end

		if ( state[DE_ASSERT_ENA_IN] )
		begin
			if ( de_assert_ena_in_count >= 8'h8 )
				next[IDLE] = 1'b1;
			else
				next[DE_ASSERT_ENA_IN] = 1'b1;
		end
		if ( state[ASSERT_RDY_IN] )
		begin
			if ( RDY_OUT )
				next[TRANSMIT_DATA] = 1'b1;
			else
				next[ASSERT_RDY_IN] = 1'b1;
		end

		if ( state[TRANSMIT_DATA] )
		begin
			if ( transmit_data_count >= 8'h8 )
				next[DE_ASSERT_RDY_IN] = 1'b1;
			else
				next[TRANSMIT_DATA] = 1'b1;
		end

		if ( state[DE_ASSERT_RDY_IN] )
		begin
//			if ( de_assert_rdy_in_count >= 8'h8 )
			if ( !RDY_OUT & rdy_out_reg )
				next[IDLE] = 1'b1;
			else
				next[DE_ASSERT_RDY_IN] = 1'b1;
		end

	end
//=================================================================
// For Simumlation, display the state names
//
//	parameter	IDLE					= 0,
//			ASSERT_ENA_IN			= 1,
//			READ_DATA				= 2,
//			DE_ASSERT_ENA_IN			= 3,
//			ASSERT_RDY_IN			= 4,
//			TRANSMIT_DATA			= 5,
//			DE_ASSERT_RDY_IN			= 6;
//
//=================================================================

`ifdef SIM
always @( posedge CLK )
begin
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << ASSERT_ENA_IN))
		state_name = "ASSERT_ENA_IN";		
	if ( state == (1 << READ_DATA))
		state_name = "READ_DATA";		
	if ( state == (1 << DE_ASSERT_ENA_IN))
		state_name = "DE_ASSERT_ENA_IN";		
	if ( state == (1 << ASSERT_RDY_IN))
		state_name = "ASSERT_RDY_IN";		
	if ( state == (1 << TRANSMIT_DATA))
		state_name = "TRANSMIT_DATA";		
	if ( state == (1 << DE_ASSERT_RDY_IN))
		state_name = "DE_ASSERT_RDY_IN";		
end
`endif		


//==============================================================================
//	Instantiate the Serial I2C Communications
//==============================================================================

	i2c_slave_comms_model               I2C_SLAVE_COMMS_MODEL_INST
	(
		.CLK                            (CLK),
		.RESET                          (RESET),

		.DEVICE_ADDRESS                 (ADDRESS),

		.I2C_DATA                       (DATA),
		.I2C_CLOCK                      (CLOCK),

		.SERIAL_I2C_TRANSMIT_DATA_IN    (transmit_data_in),
		.SERIAL_I2C_TRANSMIT_ENA_OUT    (ENA_IN),
		.SERIAL_I2C_TRANSMIT_RDY_IN     (RDY_OUT),
		
		.SERIAL_I2C_RECEIVE_DATA_OUT    (receive_data_out),
		.SERIAL_I2C_RECEIVE_ENA_IN      (ENA_OUT),
		.SERIAL_I2C_RECEIVE_RDY_OUT     (RDY_IN)
		);

endmodule

