/******************************************************
 * Title: ft220x_model.v 
 * Author: Richard Jolly                           
 * Creation Date: 11/14/2023                           
 * Description: This Module is models the FT 220X chip
 * and its FT1248 Interface protocol. It uses a 10MHz SPI interface
 *
 * Copyright Earth People Technology, Inc. 2023
 *
******************************************************/
`include "../src/define.v"

`timescale 1 ns / 10 ps

module ft220x_model
  (
	input wire                  CLK,
	input wire                  RST_N,

    // Device Interface	
	input wire                  DEVICE_CLK,
	inout wire                  DEVICE_MOSI,
	output wire                 DEVICE_MISO,
	input wire                  DEVICE_SS

    // Host Interface	
	output wire                 HOST_CLK,
	output wire                 HOST_MOSI,
	input wire                  HOST_MISO,
	output wire                 HOST_SS
	

   );


   //----------------------------------------------
   // Parameter Declarations
   //----------------------------------------------
   parameter 	IDLE                        = 0,
                RECEIVE_WORD                = 1,
                STORE_DATA_WORD             = 2,
                CHECK_FOR_LAST_WORD         = 3,
				READ_WORD_DELAY             = 4,
				RECEIVE_WORD_COMPLETE       = 5,
                WRITE_OUTGOING_MESSAGE      = 6,
                WAIT_FOR_WRITE_CYCLE_COMPLETE = 7,
                WRITE_CYCLE_COMPLETE        = 8,
                WAIT_DATA_BYTE_COMPLETE     = 9,
                CHECK_DATA_XMIT_COMPLETE    = 10;
   
   parameter 	PGA_EOC_DELAY_COUNT         = 16'h7f;
   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
   	reg [10:0]                   state, next;


	//SPI Slave Registers
   	reg                         spi_slave_xmt_en;
   	reg                         spi_slave_rcv_rdy;
   	wire   [7:0]                spi_slave_data;     
   	wire                        spi_slave_xmt_rdy;
   	wire                        spi_slave_rcv_en;
	reg                         spi_transmit_waiting;
	
	//PGA Registers
	reg    [7:0]                receive_pga_command_word;
	reg    [7:0]                receive_pga_data_word;
	reg                         receive_fifo_wren_reg;
	reg    [7:0]                spi_received_command_reg;
	reg                         store_command_word_reg;
	
	//Data Transfer registers
   	reg   [9:0]                 transmit_data_out;
	reg                         adc_data_xmit_msb;
	reg   [7:0]                 adc_internal_register;
	
	//PGA Internal Registers
	reg	[3:0]                   channel_0_gain;
	reg	[3:0]                   channel_1_gain;
	reg	[3:0]                   channel_2_gain;
	reg	[3:0]                   channel_3_gain;
	
	//Number of Channels
	reg  [2:0]                  selected_channel;
	reg                         num_chans_decrement_reg;
	
    //Incoming storage registers
   	reg	[7:0]                   incoming_word_count;
	reg                         incoming_word_count_reg;
	
	//Outgoing storage registers
   	reg	[7:0]                   outgoing_word_count;
	
    //TEST Output
	reg	[3:0]                   SPI_STATE;

      
`ifdef SIM
   	reg [8*30:1] state_name;
`endif
   
   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------
   
   //SPI data register, bi-directional
   assign   spi_slave_data = (state[WAIT_FOR_WRITE_CYCLE_COMPLETE] & spi_transmit_waiting) ? transmit_data_out :  8'hzz;
   
   //PGA Channel Selection
   assign   SELECTED_PGA_CHANNEL = {1'b1,selected_channel};
   
   //-----------------------------------------------
   // Store incoming byte from SPI Bus 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		spi_received_command_reg <= 1'b0;
			receive_pga_command_word <= 0;
	end
	else 
	  begin
	     if(spi_slave_rcv_en & !spi_received_command_reg)
		 begin
		    spi_received_command_reg <= 1'b1;
		    receive_pga_command_word <= spi_slave_data;
		 end
		 else if(state[IDLE])
	     begin
	  		spi_received_command_reg <= 1'b0;
	     end
	  end
	end


   //-----------------------------------------------
   // Assert the spi_transmit_waiting flag when the 
   // SPI has completed the receive cycle and has
   // data to transmit back to the Host
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		spi_transmit_waiting <= 1'b0;
	end
	else 
	  begin
	     if (state[STORE_DATA_WORD] & (receive_pga_command_word == 8'h3a))
		 begin
	  		 spi_transmit_waiting <= 1'b1;
		 end
	     else if (state[WRITE_CYCLE_COMPLETE])
		 begin
	  		 spi_transmit_waiting <= 1'b0;
		 end
	  end
	end

   //-----------------------------------------------
   // Write Cycle (To PGA): Set the Channel to select 
   //-----------------------------------------------
   always @( negedge CLK or negedge RST_N )
   begin
      if ( !RST_N )
      begin
         selected_channel <= 3'h7;
      end
      else
      begin
         if(state[CHECK_FOR_LAST_WORD] & (receive_pga_command_word == 8'h2a) & (incoming_word_count == 8'h2))
         begin
		    case(receive_pga_data_word[3:0])
			4'h0:
               selected_channel <= 3'h1;
			4'h1:
               selected_channel <= 3'h3;
			4'h2:
               selected_channel <= 3'h7;
			4'h3:
               selected_channel <= 3'hf;
			default:
               selected_channel <= 3'h1;
			endcase
         end
      end
   end

   //-----------------------------------------------
   // Assert the USB_RXF signal to start the transmit
   // of PGA data over SPI.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		spi_slave_xmt_en <= 1'b1;
	end
	else 
	  begin
	     if (state[WRITE_OUTGOING_MESSAGE])
		 begin
	  		 spi_slave_xmt_en <= 1'b0;
		 end
	     else if (state[WRITE_CYCLE_COMPLETE])
		 begin
	  		 spi_slave_xmt_en <= 1'b1;
		 end
	  end
	end

   //-----------------------------------------------
   // Select a PGA Channel and set the gain
   //-----------------------------------------------
   always @( negedge CLK or negedge RST_N )
   begin
      if ( !RST_N )
      begin
         channel_0_gain <= 4'h0;
         channel_1_gain <= 4'h0;
         channel_2_gain <= 4'h0;
         channel_3_gain <= 4'h0;
      end
      else
      begin
         if(state[CHECK_FOR_LAST_WORD] & (receive_pga_command_word == 8'h2a) & (incoming_word_count == 8'h2))
         begin
		    case(selected_channel)
			4'h0:
               channel_0_gain <= receive_pga_data_word[7:4];
			4'h1:
               channel_1_gain <= receive_pga_data_word[7:4];
			4'h2:
               channel_2_gain <= receive_pga_data_word[7:4];
			4'h3:
               channel_3_gain <= receive_pga_data_word[7:4];
			endcase
         end
      end
   end

   //-----------------------------------------------
   // Assert the USB_TXE signal to the SPI Slave Unit
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  spi_slave_rcv_rdy <= 1'b0;
	else 
	  begin
	     if (state[RECEIVE_WORD])
	       spi_slave_rcv_rdy <= 1'b1;
	     else if (state[CHECK_FOR_LAST_WORD]) 
	       spi_slave_rcv_rdy <= 1'b0;
	  end
     end // always @ (posedge CLK or negedge RST_N)   
 
	 
   //------------------------------------------------
   //  Receive: Create the incoming_word_count for the receive unit 
   //------------------------------------------------
   always @( negedge CLK or negedge RST_N )
   begin
      if ( !RST_N )
		begin
			incoming_word_count <= 8'h0;
			outgoing_word_count <= 8'h0;
			incoming_word_count_reg <= 1'b0;
			store_command_word_reg <= 1'b0;
		end
		else
		begin
		    if(state[STORE_DATA_WORD] & !store_command_word_reg)
			begin
			    store_command_word_reg <= 1'b1;
			    case(receive_pga_command_word)
				   8'h2a:
				      incoming_word_count <= 8'h1;
				   8'h3a:
				   begin
			          incoming_word_count <= 8'h1;
					  outgoing_word_count <= 8'h2;
					  transmit_data_out <= 8'h0a;
				   end
				endcase
			end
			else if(state[READ_WORD_DELAY] & !incoming_word_count_reg)
			begin
			   incoming_word_count_reg <= 1'b1;
			   if(incoming_word_count > 8'h0)
                  incoming_word_count <= incoming_word_count - 1;
			end
			else if( state[CHECK_FOR_LAST_WORD] )
				incoming_word_count_reg <= 8'h0;
			else if( state[RECEIVE_WORD_COMPLETE] )
				store_command_word_reg <= 1'b0;
			else if(state[WRITE_OUTGOING_MESSAGE])
			begin
			   if(outgoing_word_count > 8'h0)
                  outgoing_word_count <= outgoing_word_count - 1;
			end
		end
	end

   //------------------------------------------------
   // Read the Data from SPI Slave unit
   //------------------------------------------------
   always @( negedge CLK or negedge RST_N )
   begin
      if ( !RST_N )
		begin
			receive_pga_command_word <= 0;
			receive_pga_data_word <= 0;
			receive_fifo_wren_reg <= 1'b0;
		end
		else
		begin
			if(state[STORE_DATA_WORD] & !receive_fifo_wren_reg & (incoming_word_count == 8'h2))
			begin
				receive_pga_command_word <= spi_slave_data;
				receive_fifo_wren_reg <= 1'b1;
			end
			else if(state[STORE_DATA_WORD] & receive_fifo_wren_reg & (incoming_word_count == 8'h1))
			begin
				//receive_pga_data_word <= spi_slave_data;
				receive_fifo_wren_reg <= 1'b1;
			end
			else if( state[CHECK_FOR_LAST_WORD])
			begin
				receive_fifo_wren_reg <= 1'b0;
			end
		end
	end
		

   //--------------------------------------------------
   // Write State Machine
   //
   //
/*   parameter 	IDLE                        = 0,
                RECEIVE_WORD                = 1,
                STORE_DATA_WORD             = 2,
                CHECK_FOR_LAST_WORD         = 3,
				READ_WORD_DELAY             = 4,
				RECEIVE_WORD_COMPLETE       = 5,
                WRITE_OUTGOING_MESSAGE      = 6,
                WAIT_FOR_WRITE_CYCLE_COMPLETE = 7,
                WRITE_CYCLE_COMPLETE        = 8,
                WAIT_DATA_BYTE_COMPLETE     = 9,
                CHECK_DATA_XMIT_COMPLETE    = 10;
*/   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state <= 12'h000;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end


   always @ ( state or spi_slave_rcv_en or PGA_CLK
	      or spi_slave_rcv_rdy or spi_slave_xmt_rdy
	      or spi_slave_xmt_en )

     begin
	next = 12'h000;

	if (state[IDLE])
	begin
       if(!spi_slave_rcv_en)
	       next[RECEIVE_WORD] = 1'b1;
       else
	       next[IDLE] = 1'b1;			
	end

	if (state[RECEIVE_WORD])
	  next[STORE_DATA_WORD] = 1'b1;

	if (state[STORE_DATA_WORD])
	begin
       if (spi_slave_rcv_rdy)
	       next[CHECK_FOR_LAST_WORD] = 1'b1;
       else
	       next[STORE_DATA_WORD] = 1'b1;
	end

	if (state[CHECK_FOR_LAST_WORD])
	begin
       if ( incoming_word_count > 4'h0 )
	       next[READ_WORD_DELAY] = 1'b1;
       else 
	       next[RECEIVE_WORD_COMPLETE] = 1'b1;			
	end

	if (state[READ_WORD_DELAY])
	begin
       if (!spi_slave_rcv_en)
          next[RECEIVE_WORD] = 1'b1;
       else 
          next[READ_WORD_DELAY] = 1'b1;			
	end

	if (state[RECEIVE_WORD_COMPLETE])
	begin
	   if(PGA_CLK)
	      next[RECEIVE_WORD_COMPLETE] = 1'b1;
	   else if(!spi_transmit_waiting)
	       next[IDLE] = 1'b1;
       else if (!spi_slave_rcv_rdy)
	       next[WRITE_OUTGOING_MESSAGE] = 1'b1;
       else
	       next[RECEIVE_WORD_COMPLETE] = 1'b1;
	end

	if (state[WRITE_OUTGOING_MESSAGE])
	  next[WAIT_FOR_WRITE_CYCLE_COMPLETE] = 1'b1;

	if (state[WAIT_FOR_WRITE_CYCLE_COMPLETE])
	  begin
	  	if(!spi_slave_xmt_rdy)
	       next[WRITE_CYCLE_COMPLETE] = 1'b1;
		else
	       next[WAIT_FOR_WRITE_CYCLE_COMPLETE] = 1'b1;			
	  end 

	if (state[WRITE_CYCLE_COMPLETE])
	  begin
	  	if(!spi_slave_xmt_rdy)
	       next[WAIT_DATA_BYTE_COMPLETE] = 1'b1;
		else
	       next[WRITE_CYCLE_COMPLETE] = 1'b1;			
	  end 

	if (state[WAIT_DATA_BYTE_COMPLETE])
	begin
		if(spi_slave_xmt_rdy)
	       next[CHECK_DATA_XMIT_COMPLETE] = 1'b1;			
		else 
	       next[WAIT_DATA_BYTE_COMPLETE] = 1'b1;	
	end

	if (state[CHECK_DATA_XMIT_COMPLETE])
	begin
        if(outgoing_word_count > 0)
	       next[WRITE_OUTGOING_MESSAGE] = 1'b1;
		else 
			next[IDLE] = 1'b1;	
	end

	
`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << RECEIVE_WORD))
	  state_name = "RECEIVE_WORD";
	else if (state == (1 << STORE_DATA_WORD))
	  state_name = "STORE_DATA_WORD";
	else if (state == (1 << CHECK_FOR_LAST_WORD))
	  state_name = "CHECK_FOR_LAST_WORD";
	else if (state == (1 << READ_WORD_DELAY))
	  state_name = "READ_WORD_DELAY";
	else if (state == (1 << RECEIVE_WORD_COMPLETE))
	  state_name = "RECEIVE_WORD_COMPLETE";
	else if (state == (1 << WRITE_OUTGOING_MESSAGE))
	  state_name = "WRITE_OUTGOING_MESSAGE";
	else if (state == (1 << WAIT_FOR_WRITE_CYCLE_COMPLETE))
	  state_name = "WAIT_FOR_WRITE_CYCLE_COMPLETE";
	else if (state == (1 << WRITE_CYCLE_COMPLETE))
	  state_name = "WRITE_CYCLE_COMPLETE";
	else if (state == (1 << WAIT_DATA_BYTE_COMPLETE))
	  state_name = "WAIT_DATA_BYTE_COMPLETE";
	else if (state == (1 << CHECK_DATA_XMIT_COMPLETE))
	  state_name = "CHECK_DATA_XMIT_COMPLETE";

	  
`endif
     end // always @ (...

   //----------------------------------------------
   // Instantiate the Leaf FSM Modules
   //----------------------------------------------
   spi_tiny_slave_model         SPI_TINY_SLAVE_MODEL_INST
     (
      .CLK                      (CLK), 
      .RST_N                    (RST_N),
      
      .SPI_DATA_OUT             (PGA_MISO),
      .SPI_CLK                  (PGA_CLK),
      .SPI_DATA_IN              (PGA_MOSI),
      .SPI_CS_N                 (PGA_CS),

	  .USB_RXF_N                (spi_slave_xmt_en),
      .USB_TXE_N                (!spi_slave_rcv_rdy),
	  .USB_DATA                 (spi_slave_data),
      
	  .USB_RD_N			        (spi_slave_xmt_rdy),
	  .USB_WR                   (spi_slave_rcv_en),
	  .USB_SI                   ()
      );


endmodule 

   