//******************************************************
//
//	Earth People Proprietary Information
//
//
//	File Name:	host_ft245_model.v
//	Author:	    R. Jolly
//	Date:		November 28, 2023
//	Revision:	A
//
//
//	Development:	Slave SPI Interface to USB using FT220X Chip
//	Application:	Cylcone 10
//	Description:	This file contains Verilog code that implements the serial
//				UART Communications. Used only during Modelsim simulations.
//				It has dependent files ft220x_Model.v
//
//
//
//	Revision History:
//
//	DATE		VERSION		DETAILS
//	11/28/23		A		Created by R Jolly
//
//
//
//
//
//
//
//
//
//
//
//******************************************************/

`timescale 1ns/1ps

module host_ft245_model
  (
	input	wire               CLK,
	input	wire               RST_N,

	output	wire               HOST_FT245_RXF_N,
	output	reg                HOST_FT245_TXE_N,
	input	wire               HOST_FT245_RD_N,
	input	wire               HOST_FT245_WR,
	input	wire               HOST_FT245_RESET_N,
	output                     HOST_FT245_PWREN_N,
	input	wire               HOST_FT245_TEST,
	inout	wire [7:0]         HOST_FT245_DATA,

	input	wire [7:0]         DATA_IN,
	output	wire [7:0]         DATA_OUT,
	input	wire               READ_FT_245,
	input	wire               WRITE_FT_245,
	input	wire               READ_FT_STATUS
 

   

   );



   //----------------------------------------------
   // Parameter Declarations                        
   //----------------------------------------------
   parameter 	IDLE                     = 0,
			READ_CYCLE_START             = 1,
			READ_CYCLE_COMPLETE          = 2,
			WRITE_CYCLE_START            = 3,
			WAIT_FOR_WR_ASSERT           = 4,
			ASSERT_TXE_N                 = 5,
			WAIT_FOR_CYCLE_END           = 6,
			WRITE_CYCLE_COMPLETE         = 7,
			READ_STATUS_BYTE             = 8;
			
	//State Machine Parameters for Write to Device
   parameter 	WRITE_DEVICE_IDLE       = 0,
			WRITE_DEVICE_ASSERT_TXE     = 1,
			WRITE_DEVICE_DATA_BYTE      = 2,
 			WRITE_DEVICE_CYCLE_COMPLETE = 3;
	

   
 
   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
	//state registers
	reg	[8:0]				   state, next;

	//Read Cycle registers
	reg                        start_read_cycle;
	reg                        start_read_cycle_reg;
	reg                        read_cycle_reg;
	reg                        read_cycle_meta_reg;
	wire                       read_cycle_rdy;
	wire                       read_cycle_complete;

	//Model Write Cycle registers
	reg                        write_cycle_reg;
	reg                        write_cycle_meta_reg;
	reg                        start_write_cycle;
	reg                        host_model_write_cycle_complete;
	reg  [7:0]                 write_transmit_byte;
	reg  [2:0]                 write_to_device_state;

	//Register the HOST_FT245_WR input
	reg                        write_enable;
	reg                        write_enable_reg;

	//Regisers for the HOST_FT245_TXE_N
	reg                        txe_delay_reg;
	reg	[7:0]                  txe_delay_count;
	reg                        host_ft245_txe_n_reg;
	reg	[7:0]                  host_ft245_txe_n_count;


	//write_byte delay registers
	reg                        write_byte_control_reg;
	reg	[7:0]                  write_byte_control_count;

	//Data hold registers
	wire [7:0]                 model_read_host_data_byte;
	wire [7:0]                 transmit_byte_storage;

	//Read status registers
	reg                        read_ft_status_reg;
	reg	[7:0]                  ft_status_byte;
	integer                    read_ft_status_cntr;
	reg                        host_byte_available;

	//Read Fifo registers
   reg [7:0]                   state_read_fifo, next_read_fifo;
   reg                         fifo_wren;
   reg                         fifo_rden;
   wire [7:0]                  fifo_dout;
   wire                        fifo_full, fifo_empty;
   wire                        fifo_afull, fifo_aempty;
   wire [2:0]                  data_count;
   wire                        flush;
   reg                         read_fifo_valid;
   reg                         write_cycle_complete_reg;

	//Test
	wire	[7:0]              temp_byte;




   
`ifdef SIM
   reg [8*26:1] state_name;
`endif

   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------

	assign	HOST_FT245_DATA = write_enable ? fifo_dout : 8'hzz;
	assign	HOST_FT245_PWREN_N = 1'b1;

	assign	temp_byte = HOST_FT245_DATA;
	assign  DATA_OUT = state[READ_STATUS_BYTE] ? ft_status_byte : transmit_byte_storage;
	assign  transmit_byte_storage = state[WRITE_CYCLE_COMPLETE] ? fifo_dout : transmit_byte_storage;


   //-----------------------------------------------
   // READ_FT_STATUS register. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	ft_status_byte <= 0;
		read_ft_status_cntr <= 0;
	end
	else
	  begin
		if ( state[READ_STATUS_BYTE] )
		begin
			ft_status_byte <= 	{1'b0,HOST_FT245_RXF_N, HOST_FT245_TXE_N,
	                                 HOST_FT245_RD_N, HOST_FT245_WR, HOST_FT245_RESET_N, 
									 HOST_FT245_PWREN_N, host_byte_available};
			if(read_ft_status_cntr < 8'h10)
			     read_ft_status_cntr <= read_ft_status_cntr + 1;
		end
		else 
		begin
	       		ft_status_byte <= 0;
				read_ft_status_cntr <= 0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)

	 
   //-----------------------------------------------
   // host_byte_available register provides a bit in the
   // status register to show that the FT245 chip has 
   // received a byte and is available for read. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	host_byte_available <= 1'b0;
	end
	else
	  begin
		if (fifo_empty)
			host_byte_available <= 1'b0;
		else  
		begin
	       	host_byte_available <= 1'b1;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)

	 
   //-----------------------------------------------
   // Read the status of the FT245 chip and write it
   // to the host output   register. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	read_ft_status_reg <= 1'b0;
	end
	else
	  begin
		if ( READ_FT_STATUS )
			read_ft_status_reg <= 1'b1;
		else 
		begin
	       		read_ft_status_reg <= 1'b0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Create a start read cycle signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	read_cycle_reg <= 0;
		read_cycle_meta_reg <= 0;
	end
	else
	  begin
		if ( READ_FT_245 )
			read_cycle_meta_reg <= 1;
		//else if ( read_cycle_meta_reg && state[READ_CYCLE_START] )
		//	read_cycle_reg <= 1;
		//else if ( read_cycle_complete || host_model_write_cycle_complete )
		//begin
	    //  read_cycle_reg <= 0;
		else
			read_cycle_meta_reg <= 0;
		//end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Create a start_read_cycle signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	start_read_cycle <= 1'b0;
	 	start_read_cycle_reg <= 1'b0;
	end
	else
	  begin
		if ( state[READ_CYCLE_START] & !start_read_cycle_reg)
		begin
			 start_read_cycle <= 1'b1;
	 	     start_read_cycle_reg <= 1'b1;
		end
		else if ( state[READ_CYCLE_START] & start_read_cycle_reg)
		begin
			 start_read_cycle <= 1'b0;
	 	     start_read_cycle_reg <= 1'b1;
		end
		else if ( state[READ_CYCLE_COMPLETE] & start_read_cycle_reg)
		begin
			 start_read_cycle <= 1'b0;
	 	     start_read_cycle_reg <= 1'b0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Host FT245 write_cycle signal. Writes bytes into
   // the FIFO for transmission to the Slave SPI Iface.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	write_cycle_reg <= 0;
		write_cycle_meta_reg <= 0;
	end
	else
	  begin
		if ( WRITE_FT_245 )
           write_cycle_meta_reg <= 1;
		else if ( write_cycle_meta_reg & state[WRITE_CYCLE_START] )
           write_cycle_reg <= 1;
		else if ( !WRITE_FT_245 & state[WRITE_CYCLE_COMPLETE]  )
		begin
           write_cycle_reg <= 0;
           write_cycle_meta_reg <= 0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // FIFO write enable. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	fifo_wren <= 1'b0;
		write_cycle_complete_reg <= 1'b0;
	end
	else
	begin
		 fifo_wren <= 1'b0;
	     if(host_model_write_cycle_complete & !write_cycle_complete_reg )
		 begin
		     fifo_wren <= 1'b1;
		     write_cycle_complete_reg <= 1'b1;
		 end
		 else if(!host_model_write_cycle_complete & write_cycle_complete_reg )
		     write_cycle_complete_reg <= 1'b0;
	 end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Create a start_write_cycle signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	 	start_write_cycle <= 0;
	end
	else
	  begin
		if ( state[WRITE_CYCLE_START] )
		begin
			start_write_cycle <= 1;
			write_transmit_byte <= DATA_IN;
	    end
		else 
		begin
           write_transmit_byte <= write_transmit_byte;
           start_write_cycle <= 0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)


   //-----------------------------------------------
   // Register the WRITE_CYCLE_COMPLETE Output
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  host_model_write_cycle_complete <= 0;
	else
	  begin
	     if ( state[WAIT_FOR_CYCLE_END] )
	       host_model_write_cycle_complete <= 1;
	     else if ( state[IDLE] )
	       host_model_write_cycle_complete <= 0;
	  end 
     end // always @ (posedge CLK or negedge RST_N)
   
   //-----------------------------------------------
   // Create a txe_delay_count. This will allow
   // state[ASSERT_TXE_N] to assert for several clocks
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
				txe_delay_reg <= 0;
				txe_delay_count <= 0;
	end
	else
	begin
		if ( state[ASSERT_TXE_N] && !txe_delay_reg )
			txe_delay_reg <= 1;
		else if ( txe_delay_reg )
		begin
			txe_delay_reg <= 1;
			txe_delay_count <= txe_delay_count + 1;
			if( txe_delay_count >= 8'hf)
			begin
				txe_delay_reg <= 0;
				txe_delay_count <= 0;
			end
		end
     end
	end

   //-----------------------------------------------
   // Create a write_byte_control_count. This will allow
   // state[WRITE_CONTROL_BYTE] to assert for several clocks
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
       write_byte_control_reg <= 0;
       write_byte_control_count <= 0;
	end
	else
	begin
		if ( state[WAIT_FOR_WR_ASSERT] && !write_byte_control_reg )
			write_byte_control_reg <= 1;
		else if ( write_byte_control_reg )
		begin
			write_byte_control_reg <= 1;
			write_byte_control_count <= write_byte_control_count + 1;
			if( write_byte_control_count >= 8'hf)
			begin
				write_byte_control_reg <= 0;
				write_byte_control_count <= 0;
			end
		end
     end
	end

   //-----------------------------------------------
   // HOST_FT245_TXE_N State Machine to send bytes 
   // from the host to the device.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
   begin
	 if (!RST_N)
	 begin
		write_enable <= 1'b0;
		HOST_FT245_TXE_N <= 1'b0;
		fifo_rden <= 1'b0;
		write_to_device_state <= WRITE_DEVICE_IDLE;
	 end
	 else
	  begin
	     case(write_to_device_state)
		 WRITE_DEVICE_IDLE:
		 begin
            if(!fifo_empty)
			begin
			   HOST_FT245_TXE_N <= 1'b1;
               write_enable <= 1'b1;
			   fifo_rden <= 1'b0;
			   write_to_device_state <= WRITE_DEVICE_ASSERT_TXE;
			end
            else 
			begin
               write_enable <= 1'b0;
			   write_to_device_state <= WRITE_DEVICE_IDLE;
			end
		 end
		 WRITE_DEVICE_ASSERT_TXE:
         begin
            HOST_FT245_TXE_N <= 1'b1;
            write_enable <= 1'b1;
			fifo_rden <= 1'b1;
            write_to_device_state <= WRITE_DEVICE_DATA_BYTE;
         end
		 WRITE_DEVICE_DATA_BYTE:
		 begin
            if(HOST_FT245_WR)
			begin
			   HOST_FT245_TXE_N <= 1'b0;
               write_enable <= 1'b1;
			   fifo_rden <= 1'b0;
			   write_to_device_state <= WRITE_DEVICE_CYCLE_COMPLETE;
			end
            else
            begin			
               write_enable <= 1'b1;
			   fifo_rden <= 1'b0;
               write_to_device_state <= WRITE_DEVICE_DATA_BYTE;
			end
		 end
		 WRITE_DEVICE_CYCLE_COMPLETE:
		 begin
            if(!HOST_FT245_WR)
			begin
			   HOST_FT245_TXE_N <= 1'b0;
               write_enable <= 1'b0;
			   fifo_rden <= 1'b0;
			   write_to_device_state <= WRITE_DEVICE_IDLE;
			end
            else
            begin			
               write_enable <= 1'b1;
               write_to_device_state <= WRITE_DEVICE_CYCLE_COMPLETE;
			end
		 end
		 default:
		 begin
            HOST_FT245_TXE_N <= 1'b0;
            write_enable <= 1'b0;
            fifo_rden <= 1'b0;
            write_to_device_state <= WRITE_DEVICE_IDLE;
         end
		 endcase
	 end
    end
	
   //--------------------------------------------------
   // Host Bus State Machine
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state <= 9'h00;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

   always @ ( /*AUTOSENSE*/state or read_cycle_complete or
		host_model_write_cycle_complete or read_cycle_reg or 
		write_cycle_reg or write_cycle_meta_reg or read_ft_status_reg or
		read_cycle_meta_reg or read_ft_status_cntr or
		write_enable_reg or write_enable or txe_delay_count or 
		write_byte_control_count or start_write_cycle)
     begin
	next = 9'h00;

	if (state[IDLE])
	begin
		if( read_cycle_meta_reg )
	  		next[READ_CYCLE_START] = 1'b1;
		else if( write_cycle_meta_reg )
	  		next[WRITE_CYCLE_START] = 1'b1;
		else if(read_ft_status_reg)
	  		next[READ_STATUS_BYTE] = 1'b1;
		else
			next[IDLE] = 1'b1;
	end

	if (state[READ_CYCLE_START])
	begin
		if( read_cycle_complete )
	  		next[READ_CYCLE_COMPLETE] = 1'b1; 
		else
	  		next[READ_CYCLE_START] = 1'b1; 
	end

	if (state[READ_CYCLE_COMPLETE])
	begin
		if( !read_cycle_meta_reg )
	  		next[IDLE] = 1'b1; 
		else
	  		next[READ_CYCLE_COMPLETE] = 1'b1; 
	end

	if (state[WRITE_CYCLE_START])
	  		next[WAIT_FOR_WR_ASSERT] = 1'b1; 

	if (state[WAIT_FOR_WR_ASSERT])
	  begin
		if ( write_byte_control_count > 8'h2 )
			next[ASSERT_TXE_N] = 1'b1;
		else 
			next[WAIT_FOR_WR_ASSERT] = 1'b1;
	  end

	if (state[ASSERT_TXE_N])
	  begin
	     if ( txe_delay_count > 8'h2 )
	       next[WAIT_FOR_CYCLE_END] = 1'b1;
	     else
	       next[ASSERT_TXE_N] = 1'b1;
	  end

	if (state[WAIT_FOR_CYCLE_END])
	  begin
	     if (!start_write_cycle)
	       next[WRITE_CYCLE_COMPLETE] = 1'b1;
	     else
	       next[WAIT_FOR_CYCLE_END] = 1'b1;
	  end

	if (state[WRITE_CYCLE_COMPLETE])
	begin
		if( !write_cycle_meta_reg )
	  		next[IDLE] = 1'b1; 
		else
	  		next[WRITE_CYCLE_COMPLETE] = 1'b1; 
	end

	if (state[READ_STATUS_BYTE])
	begin
		if( read_ft_status_cntr >= 8'h0f )
	  		next[IDLE] = 1'b1; 
		else
	  		next[READ_STATUS_BYTE] = 1'b1; 
	end

	
`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << READ_CYCLE_START))
	  state_name = "READ_CYCLE_START";
	else if (state == (1 << READ_CYCLE_COMPLETE))
	  state_name = "READ_CYCLE_COMPLETE";
	else if (state == (1 << WRITE_CYCLE_START))
	  state_name = "WRITE_CYCLE_START";
	else if (state == (1 << WAIT_FOR_WR_ASSERT))
	  state_name = "WAIT_FOR_WR_ASSERT";
	else if (state == (1 << ASSERT_TXE_N))
	  state_name = "ASSERT_TXE_N";
	else if (state == (1 << WAIT_FOR_CYCLE_END))
	  state_name = "WAIT_FOR_CYCLE_END";
	else if (state == (1 << WRITE_CYCLE_COMPLETE))
	  state_name = "WRITE_CYCLE_COMPLETE";
	else if (state == (1 << READ_STATUS_BYTE))
	  state_name = "READ_STATUS_BYTE";
`endif
     end // always @ (...





   //-----------------------------------------------
   // Instantiate Host Bus Model Read
   //-----------------------------------------------
   host_ft245_model_read HOST_READ
	(
	.CLK                                 (CLK),
    .RST_N                               (RST_N),
      

	.HOST_FT245_RXF_N                    (HOST_FT245_RXF_N),
	.HOST_FT245_RD_N                     (HOST_FT245_RD_N),
	.HOST_FT245_READ_BYTE                (model_read_host_data_byte),

	.TRANSMIT_BYTE                       (DATA_IN),
	.START_READ_CYCLE                    (start_read_cycle),
	.READ_CYCLE_RDY                      (read_cycle_rdy),
	.READ_CYCLE_COMPLETE                 (read_cycle_complete)

	);



   //-----------------------------------------------
   // Instantiate Host Bus Model Write
   //-----------------------------------------------
/*   host_ft245_model_write HOST_WRITE
	(
    .CLK                             (CLK),
    .RST_N                           (RST_N),
      
	
	.HOST_FT245_TXE_N                (HOST_FT245_TXE_N),
	.HOST_FT245_WR                   (HOST_FT245_WR),
	.HOST_FT245_WRITE_BYTE           (HOST_FT245_DATA ),

	.TRANSMIT_BYTE                   (write_transmit_byte),
	.START_WRITE_CYCLE               (start_write_cycle),
	.WRITE_CYCLE_COMPLETE            (host_model_write_cycle_complete)


	);
*/
   //-----------------------------------------------
   // Sync FIFO to store bytes from the Device
   //-----------------------------------------------
  sync_fifo #(
      .ADDR_WIDTH               (10),
	  .DATA_WIDTH               (8),
	  .DEPTH                    (1024),
	  .AEMPTY                   (10),
	  .AFULL                    (1000)
	  ) FIFO
     (
	  .clk                     (CLK),
	  .reset_n                 (RST_N),
	  .flush                   (1'b0),
	  .write_data              (write_transmit_byte),	  
	  .read_req                (fifo_rden),
	  .wdata_valid             (fifo_wren),
	  .read_data               (fifo_dout),
	  .rdata_valid             (),
	  .fifo_empty              (fifo_empty),
	  .fifo_aempty             (fifo_aempty),
	  .fifo_full               (fifo_full),
	  .fifo_afull              (fifo_afull),
	  .write_ack               ()
      );


   

endmodule // 

