//================================================================
//
//	Earth People Proprietary Information
//
//
//	File Name:	slave_spi_comms_model.v
//	Author:	    R. Jolly
//	Date:		November 28, 2023
//	Revision:	A
//
//
//	Development:	Slave SPI Interface to USB using FT220X Chip
//	Application:	Cylcone 10
//	Description:	This file contains Verilog code that implements the serial
//				UART Communications. Used only during Modelsim simulations.
//				It has dependent files ft220x_Model.v
//
//
//
//	Revision History:
//
//	DATE		VERSION		DETAILS
//	11/28/23		A		Created by R Jolly
//
//
//
//
//
//
//
//
//
//
//
//==================================================================

`include "../Testbench/tb_define.v"


module slave_spi_comms_model (		
	input  wire                 CLK,
	input  wire                 RESET,

	input  wire                 SPI_CLK,
	input  wire                 MOSI,
	output wire                 MISO,
	inout  wire                 MIOSIO_1,
	input  wire                 SPI_SS,

	input  wire                 FT245_RXF_N,
	input  wire                 FT245_TXE_N,
	output wire                 FT245_RD_N,
	output reg                  FT245_WR,
	output wire                 FT245_RESET_N,
	input  wire                 FT245_PWREN_N,
	input  wire                 FT245_TEST,
	inout  wire [7:0]           FT245_DATA,

	output reg  [7:0]           CONTROL_REGISTER,
	input  wire [7:0]           STATUS_REGISTER,
	output wire [7:0]           RECEIVE_REGISTER,
	input  wire [7:0]           SEND_REGISTER
	
		
		);

//==================================================================
// Parameter Declarations
//==================================================================
	parameter IDLE					    = 0,
			WAIT_FOR_CMD				= 1,
			WRITE_CMD_NOP				= 2,
			CHECK_FOR_READ_WRITE_CMD    = 3,
			WRITE_CYCLE_START           = 4,
			WRITE_BYTE                  = 5,
			CHECK_FOR_WRITE_BYTE_COMPLETE = 6,
			READ_CYCLE_START            = 7,
			READ_BYTE                   = 8,
			CHECK_FOR_READ_BYTE_COMPLETE = 9;
		
	//State Machine Parameters for Write to Device
   parameter 	RECIEVE_HOST_IDLE       = 0,
			RECIEVE_HOST_ASSERT_TXE     = 1,
			RECIEVE_HOST_DATA_BYTE      = 2,
 			RECIEVE_HOST_CYCLE_COMPLETE = 3;

//==================================================================
// Register Declarations
//==================================================================

	// CONTROL Register
	reg                        control_update_reg;
	reg                        status_update_reg;

	// State Machine Registers
	reg [9:0]                 state, next;


   	reg                        write_tx_byte;
   	reg [2:0]                  tx_byte_count;
   	reg [2:0]                  total_tx_bytes;
   	reg [8:0]                  bytes_written_reg;
//   	reg [3:0]                  STATE_OUT;
   
   	wire                       tx_buffer_full;
   	wire                       tx_buffer_empty;

	// Transmit Registers
	reg                        transmit_data_ready;
	reg                        transmit_data_ready_reg;
	reg                        cmd_reg;

	// Receive Registers
	reg                        receive_data_ready;
	reg                        receive_data_ready_reg;


	//FT220X Interface Registers
	reg                        host_cmd_byte;
	reg                        host_cmd_reg;
	
	//FT 245 Registers
	wire                       host_ft245_rxf_n;
	wire                       host_ft245_txe_n;
	wire [7:0]                 host_ft245_data;
	wire                       host_ft245_rd_n;
	wire                       host_ft245_wr;	

	//SPI Read/Write Command Registers
	reg                        status_cmd;
	reg                        read_cmd;
	reg                        write_cmd;
	reg                        write_byte_fifo_empty;
	reg                        read_byte_fifo_empty;
	
	//SPI Registers
   	reg  [7:0]                 spi_data_byte_write;
	reg                        spi_write_enable;	  
   	wire [7:0]                 spi_data_byte_read;
	wire                       spi_read_enable;	  
   	reg  [7:0]                 spi_command_byte;
	reg                        spi_read_ready;
	wire                       spi_write_ready;
	
	//FT245_TXE_N State Machine
	reg                        host_data_ready;
	reg  [2:0]                 recieve_from_host_state;
	
	//MIOSIO Direction control registers
	reg                        write_to_miosio;
	reg                        write_to_miosio_reg;
	wire                       miosio_mosi;
	
	//MOSI Status output registers
	reg                        write_to_miso;
	reg                        write_to_miso_reg;
	reg                        miso_write_status;
	wire                       miso_from_slave_spi_iface;
	
	
	


`ifdef SIM
   reg [8*25:1] state_name;
`endif
   
//==================================================================
// Signal Assignments  
//==================================================================
	assign 	 tx_buffer_full = !tx_buffer_empty;

	assign	RECEIVE_REGISTER = 8'h09;

	assign	host_ft245_rxf_n = FT245_RXF_N;
	assign	FT245_RD_N = host_ft245_rd_n;
	assign	FT245_RESET_N = 1'b0;
	assign	host_ft245_data = FT245_WR ? FT245_DATA : host_ft245_data;
	assign	FT245_DATA = spi_read_enable ? spi_data_byte_read : 8'hzz;
	assign  miosio_mosi = MIOSIO_1;
	assign  MIOSIO_1 = write_to_miosio ? miso_from_slave_spi_iface/*MISO*/ : 1'bz;
	assign  MISO = write_to_miso ? miso_write_status : 1'b1/*MISO*/;

//==================================================================
// Create a Tx State Test Output Signal
//==================================================================
//   always @(posedge CLK or negedge RESET)
//     begin
//	if (!RESET)
//	  STATE_OUT <= 0;
//	else
//	  begin
//	     if (state[IDLE])
//               STATE_OUT <= 4'h0;
//	     if (state[WAIT_FOR_CMD])
//	       STATE_OUT <= 4'h1 ;
//	     if (state[CMD_TX_WRITE])
//	       STATE_OUT <= 4'h2 ;
//	     if (state[WRITE_CMD_NOP])
//	       STATE_OUT <= 4'h3 ;
//	     if (state[WR_CMD_WAIT_BUF_EMPTY])
//	       STATE_OUT <= 4'h4 ;
//	     if (state[DATA_TX_WRITE])
//	       STATE_OUT <= 4'h5 ;
//	     if (state[WRITE_CMD_NOP])
//	       STATE_OUT <= 4'h6 ;
//	  end // else: !if(!RST_N)
//     end // always @ (posedge CLK or negedge RST_N)



//=================================================================================
//	Receive Cycle: CONTROL_REGISTER, this register informs the ARM9 when sensor
//	data has been updated. uart_rdn
//=================================================================================

/*
	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			CONTROL_REGISTER <= 8'b0;
			control_update_reg <= 1'b0;
			status_update_reg <= 1'b0;
		end
		else
		begin
			if( STATUS_REGISTER[3] && !status_update_reg )
			begin
				CONTROL_REGISTER[3] <= 1'b0;
				status_update_reg <= 1'b1;
			end
			else if( STATUS_REGISTER[3]&& status_update_reg )
			begin
				status_update_reg <= 1'b1;
			end
			else if( !STATUS_REGISTER[3] && status_update_reg )
			begin
				status_update_reg <= 1'b0;
			end

			if ( state[RECEIVE_BYTE] & !control_update_reg )
			begin
				CONTROL_REGISTER[3] <= CONTROL_REGISTER[3] || 1'b1;
				control_update_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				control_update_reg <= 1'b0;
			end
		end
	end

*/

      
//==================================================================
// Transmit: transmit_data_ready will be equated with STATUS_REGISTER[1]
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			transmit_data_ready <= 0;
			transmit_data_ready_reg <= 0;
		end
		else 
		begin
			if ( STATUS_REGISTER[1] && !transmit_data_ready_reg )
			begin
				transmit_data_ready <= 1'b1;
				transmit_data_ready_reg <= 1'b1;
			end
			else if ( STATUS_REGISTER[1] && transmit_data_ready_reg )
			begin
				transmit_data_ready <= 1'b0;
				transmit_data_ready_reg <= 1'b1;
			end
			else if ( state[IDLE] )
			begin
				transmit_data_ready <= 0;
				transmit_data_ready_reg <= 0;
			end
		end		
	end // always @ (posedge CLK or negedge RESET)

   //-----------------------------------------------
   // Register to control the direction of the MIOSIO_1 
   // signal. During the Write cycle, the Slave SPI
   // will write to the Master. During Read cycles,
   // the Slave SPI will read from Master.
   //-----------------------------------------------
   always @(posedge CLK or negedge RESET)
   begin
	 if (!RESET)
	 begin
		write_to_miosio <= 1'b0;
		write_to_miosio_reg <= 1'b0;
	 end
	 else
	 begin
        if(next[WRITE_CYCLE_START] & read_cmd & !write_to_miosio_reg)
        begin
		   write_to_miosio <= 1'b1;
		   write_to_miosio_reg <= 1'b1;
		end
		else if(SPI_SS & write_to_miosio_reg)
        begin
		   write_to_miosio <= 1'b0;
		   write_to_miosio_reg <= 1'b0;
		end
	 end
   end
   
   //-----------------------------------------------
   // Register to control the MISO signal. 
   // During the Write cycle, the Slave SPI
   // will write the status of the RXE signal 
   // to the Master. 
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RESET)
   begin
	 if (!RESET)
	 begin
		write_to_miso <= 1'b0;
		write_to_miso_reg <= 1'b0;
		miso_write_status <= 1'b1;
	 end
	 else
	 begin
        if(next[WRITE_CYCLE_START] & read_cmd & !write_to_miso_reg)
        begin
		   write_to_miso <= 1'b1;
		   write_to_miso_reg <= 1'b1;
		   if(host_data_ready)
		      miso_write_status <= 1'b0;
		   else
		      miso_write_status <= 1'b1;
		end
		else if(SPI_SS & write_to_miso_reg)
        begin
		   write_to_miso <= 1'b0;
		   write_to_miso_reg <= 1'b0;
		   miso_write_status <= 1'b1;
		end
	 end
   end
   
   //-----------------------------------------------
   // FT245_WR State Machine to recieve bytes 
   // from the host to the device.
   //-----------------------------------------------
   always @(posedge CLK or negedge RESET)
   begin
	 if (!RESET)
	 begin
		host_data_ready <= 1'b0;
		FT245_WR <= 1'b0;
		recieve_from_host_state <= RECIEVE_HOST_IDLE;
	 end
	 else
	  begin
	     case(recieve_from_host_state)
		 RECIEVE_HOST_IDLE:
		 begin
            if(FT245_TXE_N)
			begin
			   FT245_WR <= 1'b1;
               host_data_ready <= 1'b1;
			   recieve_from_host_state <= RECIEVE_HOST_ASSERT_TXE;
			end
            else 
			begin
               host_data_ready <= 1'b0;
			   recieve_from_host_state <= RECIEVE_HOST_IDLE;
			end
		 end
		 RECIEVE_HOST_ASSERT_TXE:
         begin
            FT245_WR <= 1'b1;
            host_data_ready <= 1'b1;
            recieve_from_host_state <= RECIEVE_HOST_DATA_BYTE;
         end
		 RECIEVE_HOST_DATA_BYTE:
		 begin
            if(!FT245_TXE_N)
			begin
			   FT245_WR <= 1'b0;
               host_data_ready <= 1'b1;
			   recieve_from_host_state <= RECIEVE_HOST_CYCLE_COMPLETE;
			end
            else
            begin			
               host_data_ready <= 1'b1;
               recieve_from_host_state <= RECIEVE_HOST_DATA_BYTE;
			end
		 end
		 RECIEVE_HOST_CYCLE_COMPLETE:
		 begin
            if(state[CHECK_FOR_WRITE_BYTE_COMPLETE])
			begin
			   FT245_WR  <= 1'b0;
               host_data_ready <= 1'b0;
			   recieve_from_host_state <= RECIEVE_HOST_IDLE;
			end
            else
            begin			
               host_data_ready <= 1'b1;
               recieve_from_host_state <= RECIEVE_HOST_CYCLE_COMPLETE;
			end
		 end
		 default:
		 begin
            FT245_WR <= 1'b0;
            host_data_ready <= 1'b0;
            recieve_from_host_state <= RECIEVE_HOST_IDLE;
         end
		 endcase
	 end
    end
	
	
//==================================================================
// SPI Command and Data Phase
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			host_cmd_byte <= 1'b0;
			host_cmd_reg <= 1'b0;
		end
		else
		begin
			if(!SPI_SS & !host_cmd_reg)
			begin
				host_cmd_byte <= 1'b1;
				host_cmd_reg <= 1'b1;
			end
			else if (state[WRITE_CMD_NOP] & host_cmd_reg)
			begin
               host_cmd_byte <= 1'b0;
			   host_cmd_reg <= 1'b1;
			end
			else if (SPI_SS & host_cmd_reg)
			begin
               host_cmd_byte <= 1'b0;
			   host_cmd_reg <= 1'b0;
			end
		end 
	end // always @ (posedge CLK or negedge RESET)

//==================================================================
// SPI Recieve Data Byte
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			spi_read_ready <= 1'b1;
			spi_write_enable <= 1'b0;
			spi_command_byte <= 0;
	    end
		else
		begin
			if(spi_read_enable & state[WAIT_FOR_CMD])
			begin
				spi_read_ready <= 1'b1;
				spi_command_byte <= spi_data_byte_read;
			end
			else if (state[CHECK_FOR_READ_WRITE_CMD])
			begin
				if(read_cmd)
				begin
				   spi_read_ready <= 1'b0;
				   spi_write_enable <= 1'b1;
				   if(host_data_ready)
				      spi_data_byte_write <= host_ft245_data;
				   else
				      spi_data_byte_write <= `SPI_READ_CMD_NO_DATA;
				end
				else if(status_cmd)
				begin
				   spi_read_ready <= 1'b0;
				   spi_write_enable <= 1'b1;
				   spi_data_byte_write <= `SPI_STATUS_CMD_DATA;
				end
				else if(write_cmd)
				begin
				   spi_read_ready <= 1'b1;
				   spi_write_enable <= 1'b0;
				end
			end
			else
			begin
			   spi_read_ready <= 1'b1;
			   spi_write_enable <= 1'b0;
			end
		end 
	end // always @ (posedge CLK or negedge RESET)
   
	
//==================================================================
// SPI Check Command Byte
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
		   status_cmd <= 1'b0;
           read_cmd <= 1'b0;
           write_cmd <= 1'b0;
		end
		else
		begin
			if(state[CHECK_FOR_READ_WRITE_CMD])
			begin
			   case(spi_command_byte)
			      `STATUS_CMD_IN:
				  begin
		             status_cmd <= 1'b1;
                     read_cmd <= 1'b0;
                     write_cmd <= 1'b0;
		          end
			      `WRITE_CMD_IN:
				  begin
		             status_cmd <= 1'b0;
                     read_cmd <= 1'b0;
                     write_cmd <= 1'b1;
		          end
			      `READ_CMD_IN:
				  begin
 		             status_cmd <= 1'b0;
                     read_cmd <= 1'b1;
                     write_cmd <= 1'b0;
		          end
			     default:
				  begin
		             status_cmd <= 1'b0;
                     read_cmd <= 1'b0;
                     write_cmd <= 1'b0;
		          end
			   endcase
			end
			else if(state[READ_CYCLE_START] | state[WRITE_CYCLE_START])
			begin
		             status_cmd <= 1'b0;
                     read_cmd <= 1'b0;
                     write_cmd <= 1'b0;
			end
		end 
	end // always @ (posedge CLK or negedge RESET)

//==================================================================
// SPI Determine bytes to send/recieve
//==================================================================
	always @(posedge CLK or negedge RESET)
	begin
		if (!RESET)
		begin
			write_byte_fifo_empty <= 1'b0;
			read_byte_fifo_empty <= 1'b0;
		end
		else
		begin
			if(!SPI_SS)
			begin
			   write_byte_fifo_empty <= 1'b0;
			   read_byte_fifo_empty <= 1'b0;
			end
			else if(!SPI_SS & write_cmd)
			begin
			   write_byte_fifo_empty <= 1'b1;
			   read_byte_fifo_empty <= 1'b0;
			end
			else if(!SPI_SS & (read_cmd | status_cmd))
			begin
			   write_byte_fifo_empty <= 1'b0;
			   read_byte_fifo_empty <= 1'b1;
			end
			else if(SPI_SS)
			begin
			   write_byte_fifo_empty <= 1'b0;
			   read_byte_fifo_empty <= 1'b0;
			end
			else if(host_cmd_byte)
			begin
			   write_byte_fifo_empty <= 1'b0;
			   read_byte_fifo_empty <= 1'b0;
			end
		end 
	end // always @ (posedge CLK or negedge RESET)
   
   
//==================================================================
// Finite  State Machine
//==================================================================
   // Next State Logic
   always @(posedge CLK or negedge RESET)
     begin
	if (!RESET)
	  begin
	     state <= 10'h000;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

     // State Definitions
   always @ ( state or spi_write_enable or spi_read_enable
             or read_cmd or write_cmd or write_byte_fifo_empty
			 or read_byte_fifo_empty or host_cmd_byte or spi_read_ready
			 or spi_write_ready)
     begin
	next = 10'h000;
	
	if (state[IDLE])
	begin
		if (host_cmd_byte)
			next[WAIT_FOR_CMD] = 1'b1;
		else 
			next[IDLE] = 1'b1;
	end

	if (state[WAIT_FOR_CMD])
	  begin
	     if (spi_read_enable)
	       next[WRITE_CMD_NOP] = 1'b1;
	     else
	       next[WAIT_FOR_CMD] = 1'b1;
	  end
	
	if (state[WRITE_CMD_NOP])
       if (!spi_read_enable)
          next[CHECK_FOR_READ_WRITE_CMD] = 1'b1;
	     else
	       next[WRITE_CMD_NOP] = 1'b1;

	if (state[CHECK_FOR_READ_WRITE_CMD])
    begin
	   	if(read_cmd | status_cmd)
		   next[WRITE_CYCLE_START] = 1'b1;
		else if(write_cmd)
		   next[READ_CYCLE_START] = 1'b1;
		else 
           next[CHECK_FOR_READ_WRITE_CMD] = 1'b1;	
    end		   
		   
		   
	if (state[WRITE_CYCLE_START])
    begin
		  if(spi_write_enable)
		    next[WRITE_BYTE] = 1'b1;
		  else
		    next[WRITE_CYCLE_START] = 1'b1;
    end 

	if (state[WRITE_BYTE])
    begin
		  if(!spi_write_ready)
		    next[CHECK_FOR_WRITE_BYTE_COMPLETE] = 1'b1;
		  else
		    next[WRITE_BYTE] = 1'b1;
    end 

	if (state[CHECK_FOR_WRITE_BYTE_COMPLETE])
    begin
		  if(!write_byte_fifo_empty)
		    next[IDLE] = 1'b1;
		  else
		    next[WRITE_CYCLE_START] = 1'b1;
    end 

	if (state[READ_CYCLE_START])
    begin
	     if(spi_read_ready)
	       next[READ_BYTE] = 1'b1;
	     else
	       next[READ_CYCLE_START] = 1'b1;
    end

		if (state[READ_BYTE])
    begin
		  if(spi_read_ready)
		    next[CHECK_FOR_READ_BYTE_COMPLETE] = 1'b1;
		  else
		    next[READ_BYTE] = 1'b1;
    end 

	if (state[CHECK_FOR_READ_BYTE_COMPLETE])
    begin
		  if(!read_byte_fifo_empty)
		    next[IDLE] = 1'b1;
		  else
		    next[READ_CYCLE_START] = 1'b1;
    end 


//==================================================================
// For simulation, display the state names
//==================================================================
`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << WAIT_FOR_CMD))
	  state_name = "WAIT_FOR_CMD";
	else if (state == (1 << WRITE_CMD_NOP))
	  state_name = "WRITE_CMD_NOP";
	else if (state == (1 << CHECK_FOR_READ_WRITE_CMD))
	  state_name = "CHECK_FOR_READ_WRITE_CMD";
	else if (state == (1 << WRITE_CYCLE_START))
	  state_name = "WRITE_CYCLE_START";
	else if (state == (1 << WRITE_BYTE))
	  state_name = "WRITE_BYTE";
	else if (state == (1 << CHECK_FOR_WRITE_BYTE_COMPLETE))
	  state_name = "CHECK_FOR_WRITE_BYTE_COMPLETE";
	else if (state == (1 << READ_CYCLE_START))
	  state_name = "READ_CYCLE_START";
	else if (state == (1 << READ_BYTE))
	  state_name = "READ_BYTE";
	else if (state == (1 << CHECK_FOR_READ_BYTE_COMPLETE))
	  state_name = "CHECK_FOR_READ_BYTE_COMPLETE";
`endif
     end // always @ (...
   
//==============================================================================
//	Instantiate the UART Interface
//==============================================================================

	slave_spi_iface_model		SPI_IFACE_INST
	(
      .CLK                      (CLK), 
      .RST_N                    (RESET),
      
      .SPI_DATA_OUT             (miso_from_slave_spi_iface),//(MISO),
      .SPI_CLK                  (SPI_CLK),
      .SPI_DATA_IN              (miosio_mosi),//(MOSI),
      .SPI_CS_N                 (SPI_SS),

	  .DIN                      (spi_data_byte_write),      
	  .WE_IN			        (spi_write_enable),	
      .WE_RDY_OUT               (spi_write_ready),	  

	  .DOUT                     (spi_data_byte_read),
      .RE_OUT_N                 (spi_read_enable),	  
	  .RE_RDY_OUT               (spi_read_ready)
	  
		);

endmodule 
