/******************************************************
 * Title: spi_iface_model.v 
 * Author: Richard Jolly                           
 * Creation Date: 1/15/2023                           
 * Description: This Module is does serial byte writes
 * on the SPI TX line. The input is parallel data
 *
 * Copyright Earth People Technology, Inc. 2023
 *
******************************************************/

`timescale 1 ns / 10 ps

module slave_spi_iface_model
  (
	input wire                  CLK,
	input wire                  RST_N,

  
	output reg                  SPI_DATA_OUT, //To external SPI bus
	input wire                  SPI_CLK, //From external SPI bus
    input wire                  SPI_DATA_IN, //From external SPI bus
	input wire                  SPI_CS_N,  //From external SPI bus
  
	input  wire [7:0]           DIN, //Slave SPI Write Cycle
	input  wire                 WE_IN,
   	output reg                  WE_RDY_OUT,

  	output reg [7:0]            DOUT,  //Slave SPI Read Cycle
   	output reg                  RE_OUT_N,
	input  wire                 RE_RDY_OUT


   );


   //----------------------------------------------
   // Parameter Declarations                        
   //----------------------------------------------
   parameter IDLE                   = 0,
	     RECEIVE_DATA				= 1,
		 SEND_TO_HOST			    = 2,
		 RECEIVE_COMPLETE			= 3;
		 
   parameter IDLE_TRANSMIT	        = 0,
		 WRITE_REGISTER             = 1,
		 LOAD_SHIFT_REG				= 2,
		 DELAY_WRITE_DATA_BYTE      = 3,
	     WRITE_DATA_BYTE			= 4,
	     WRITE_DATA_COMPLETE		= 5;
		 
   //----------------------------------------------
   // Bit Count Parameter Declarations                        
   //----------------------------------------------
   parameter BIT_COUNT_IDLE         = 0,
			 BIT_COUNT_CLK_HI		= 1,
			 BIT_COUNT_CLK_LOW		= 2,
			 BIT_COUNT_INC_BYTE		= 3;
			 
   //parameter WE_RDY_DELAY_COUNT   = 16'h4d;
   parameter WE_RDY_DELAY_COUNT   = 16'h19;

   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
   	reg [3:0]                 state_receive, next_receive;
   	reg [5:0]                 state_transmit, next_transmit;


    // Bit Count Registers
	reg [3:0]                  in_bit_count;
	reg [3:0]                  out_bit_count;
	
	// SPI Shift Registers
	reg [7:0]                  shift_in_reg, shift_out_reg;
	reg                        read_spi_data_in_once_flag;
	reg                        write_spi_data_out_once_flag;

	reg                        write_spi_data_out_response_once;

	reg                        load_shift_reg_high_byte;

	reg                        spi_clk_reg;
	
	// Storage Registers
	reg [7:0]                  transmit_byte_storage;
	reg [2:0]                  store_reg_state; 
	
	
	// in bit counter
	reg [1:0]                  bit_count_state;
	
	//Metastable Registers
	reg                        spi_test_reg;
	reg                        we_in_reg;
	
	//Write Enable registers
	reg                        we_rdy_reg;
	reg [15:0]                 we_rdy_counter;
	
	//SPI Incoming byte detect
	reg                        spi_incoming_byte;  
	reg  [1:0]                 spi_incoming_reg;  
	
	

	//STATE_OUT Regisetrs
    reg [3:0]                  STATE_OUT;
   	
`ifdef SIM
   reg [8*26:1] state_name_receive;
   reg [8*26:1] state_name_transmit;
`endif


   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------
   

 
   //-----------------------------------------------
   // Create a State Test Output Signal
   //-----------------------------------------------
/*   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  STATE_OUT <= 0;
	else
	  begin
	     if (state[IDLE])
           STATE_OUT <= 4'h0;
	     if (state[RECEIVE_DATA])
	       STATE_OUT <= 4'h1;
	     if (state[SEND_TO_HOST])
	       STATE_OUT <= 4'h2;
	     if (state[RECEIVE_COMPLETE])
	       STATE_OUT <= 4'h3;
	     if (state[WRITE_REGISTER])
	       STATE_OUT <= 4'h4;
	     if (state[LOAD_SHIFT_REG])
	       STATE_OUT <= 4'h5;
	     if (state[DELAY_WRITE_DATA_BYTE])
	       STATE_OUT <= 4'h6;
	     if (state[WRITE_DATA_BYTE])
	       STATE_OUT <= 4'h7;
	     if (state[WRITE_DATA_COMPLETE])
	       STATE_OUT <= 4'h8;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)
*/


   //-----------------------------------------------
   //Register the we_in_reg input signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     we_in_reg <= 1'b0;
	else
	  begin
	     if(WE_IN & !SPI_CS_N)
	         we_in_reg <= 1'b1;
		 else if(state_transmit[WRITE_DATA_COMPLETE])
	         we_in_reg <= 1'b0;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Assert the WE_RDY_OUT input signal. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	     WE_RDY_OUT <= 1'b1;
		 we_rdy_reg <= 1'b0;
		 we_rdy_counter <= 0;
	end
	else
	  begin
	     if(state_transmit[WRITE_REGISTER])
	         WE_RDY_OUT <= 1'b1;
		 else if(state_transmit[WRITE_DATA_COMPLETE] & !we_rdy_reg)
		 begin
	         WE_RDY_OUT <= 1'b1;
			 we_rdy_reg <= 1'b1;
		 end
		 else if(we_rdy_reg)
		 begin
		     if(we_rdy_counter < WE_RDY_DELAY_COUNT)
			    we_rdy_counter <= we_rdy_counter + 1'd1;
			 else
	         begin
	            WE_RDY_OUT <= 1'b0;
	            we_rdy_reg <= 1'b0;
	            we_rdy_counter <= 0;
	         end
		 end
		 else if(state_transmit[IDLE_TRANSMIT])
         begin
	            WE_RDY_OUT <= 1'b1;
	            we_rdy_reg <= 1'b0;
	            we_rdy_counter <= 0;
         end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Slave SPI Receive Cycle - Sending byte to Calling
   // layer.   
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	   DOUT <= 0;
	   RE_OUT_N <= 1'b0;
	end
	else
	  begin
	     if (state_receive[SEND_TO_HOST] & RE_RDY_OUT )
		 begin
	         DOUT <= shift_in_reg;
			 RE_OUT_N <= 1'b1;
	     end
		 else
	         RE_OUT_N <= 1'b0;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Slave SPI Receive Cycle - Start Receive State
   // Machine. First byte after SPI_CS_N goes low
   // is a command byte   
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
       spi_incoming_byte <= 1'b0;  
       spi_incoming_reg <= 2'b00;  
	end
	else
	  begin
	     if(!SPI_CS_N & (spi_incoming_reg == 2'b00))
		 begin
            spi_incoming_byte <= 1'b1;
            spi_incoming_reg <= 2'b01;
	     end
		 else if((in_bit_count == 1) & (spi_incoming_reg == 2'b01))
		 begin
            spi_incoming_byte <= 1'b0;
			spi_incoming_reg <= 2'b10;
		 end
		 else if((in_bit_count == 8) & (spi_incoming_reg == 2'b10))
		 begin
            spi_incoming_byte <= 1'b0;
			spi_incoming_reg <= 2'b11;
		 end
		 else if(RE_OUT_N & (spi_incoming_reg == 2'b11))
		 begin
            spi_incoming_byte <= 1'b0;
			spi_incoming_reg <= 2'b00;
		 end
		 else if(SPI_CS_N)
		 begin
            spi_incoming_byte <= 1'b0;
			spi_incoming_reg <= 2'b00;
	     end

      end
     end // always @ (posedge CLK or negedge RST_N)



   //-----------------------------------------------
   // Slave SPI Write Cycle - Store the byte for 
   // transmission from Host.   
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	     transmit_byte_storage <= 0;
	else
	  begin
	     if(state_transmit[WRITE_REGISTER])
	         transmit_byte_storage <= DIN;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //-----------------------------------------------
   // Register the SPI_CLK input with spi_clk_reg. 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	 spi_clk_reg <= 0;
	else
	  begin
	     if ( SPI_CLK )
	       spi_clk_reg <= 1;
	     else 
	       spi_clk_reg <= 0;
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //------------------------------------------------
   // Input bit counter, maintains the count of serial
   // bits shifted in on SPI_MOSI
   //------------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  in_bit_count <= 0;
	  bit_count_state <= BIT_COUNT_IDLE;
	end
	else
	  begin
	  
	     case( bit_count_state )
		     BIT_COUNT_IDLE:
			     if ( state_receive[RECEIVE_DATA] )
				     bit_count_state <= BIT_COUNT_CLK_HI;
			 BIT_COUNT_CLK_HI:
			     if( spi_clk_reg )
				 begin
				     bit_count_state <= BIT_COUNT_CLK_LOW;
	                 in_bit_count <= in_bit_count + 1;
				 end
			 BIT_COUNT_CLK_LOW:
			     if( !spi_clk_reg )
				     //bit_count_state <= BIT_COUNT_IDLE;
				     bit_count_state <= BIT_COUNT_CLK_HI;
				 else if( in_bit_count >= 8'h8)
				     bit_count_state <= BIT_COUNT_INC_BYTE;
			 BIT_COUNT_INC_BYTE:
			     if( state_receive[RECEIVE_COMPLETE] )
				 begin
				     bit_count_state <= BIT_COUNT_IDLE;
	                 in_bit_count <= 0;
				 end
			 default:
			 begin
				     bit_count_state <= BIT_COUNT_IDLE;
			 end
		 endcase
	  end
     end

   //------------------------------------------------
   // Output bit counter keeps track of which bit is 
   // being transmitted
   //------------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
    begin
	   out_bit_count <= 0;
        write_spi_data_out_once_flag <= 1'b1;
	end
	else
	  begin
	     if(state_transmit[WRITE_DATA_BYTE] & !spi_clk_reg & write_spi_data_out_once_flag )
	     begin
		    //out_bit_count <= out_bit_count + 1;
			write_spi_data_out_once_flag <= 0; 
	     end
		else if(state_transmit[WRITE_DATA_BYTE] & spi_clk_reg & !write_spi_data_out_once_flag )
 		begin 
	       	out_bit_count <= out_bit_count + 1;
        	write_spi_data_out_once_flag <= 1;
		end
	    else if(state_transmit[WRITE_DATA_COMPLETE] )
		begin
	        out_bit_count <= 0;
		end
	  end
     end // always @ (posedge CLK or negedge RST_N)

   //------------------------------------------------
   // SPI input shift register
   //------------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
			shift_in_reg <= 0;
			read_spi_data_in_once_flag <= 1;
	end
	else
	  begin
	     if ( state_receive[RECEIVE_DATA] & spi_clk_reg & read_spi_data_in_once_flag )
	       begin
		  shift_in_reg[0] <= SPI_DATA_IN;
		  shift_in_reg[7:1] <= shift_in_reg[6:0];
		  read_spi_data_in_once_flag <= 0;
	       end
		else if ( state_receive[RECEIVE_DATA] & !spi_clk_reg & !read_spi_data_in_once_flag )
		begin
			read_spi_data_in_once_flag <= 1;
		end
		else if (state_transmit[WRITE_REGISTER])
			read_spi_data_in_once_flag <= 1;
	     else if (state_receive[IDLE])
	       shift_in_reg <= 0;
	  end
     end // always @ (posedge CLK or negedge RST_N)
 
   //------------------------------------------------
   // SPI Shift Data Response out
   //------------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     shift_out_reg <= 8'h30;
	     SPI_DATA_OUT <= 1;
		 load_shift_reg_high_byte <= 0;
		 write_spi_data_out_response_once <= 0;
	  end
	else
	  begin

	     if (state_transmit[LOAD_SHIFT_REG] )
	     begin
	  		SPI_DATA_OUT <= 0;
			shift_out_reg <= transmit_byte_storage;
	     end
// Added new shift_out_reg <= register_value code
	     //if (state_receive[RECEIVE_DATA])
	     //begin
		 //    SPI_DATA_OUT <= SPI_DATA_OUT;
		 //    shift_out_reg <= 0;
	     //end
	     else if (state_transmit[IDLE_TRANSMIT])
	     begin
	  		SPI_DATA_OUT <= 0;
	     end
	     else if ( state_transmit[WRITE_DATA_BYTE] )
	       begin

			if(!spi_clk_reg & !write_spi_data_out_response_once & !write_spi_data_out_once_flag)
			begin
		  		SPI_DATA_OUT <= shift_out_reg[7];
		  		shift_out_reg[7:1] <= shift_out_reg[6:0];
		 		shift_out_reg[0] <= 0;
				write_spi_data_out_response_once <= 1'b1;
			end
			else if(spi_clk_reg & write_spi_data_out_response_once & write_spi_data_out_once_flag)
			begin
				write_spi_data_out_response_once <= 1'b0;
			end

	       end // if ((state[WRITE_BUSY_RESPONSE] || state[WRITE_CMD_RESPONSE]) &&...
	  end
     end // always @ (posedge CLK or negedge RST_N)   
   
	  
   //--------------------------------------------------
   // SPI State Machine
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state_receive <= 4'h00;
	     state_receive[IDLE] <= 1'b1;
	  end
	else
	  state_receive <= next_receive;
     end

   always @ ( /*AUTOSENSE*/state_receive
         or SPI_CLK or SPI_DATA_IN or SPI_CS_N
	     or in_bit_count or spi_clk_reg or store_reg_state 
		 or bit_count_state or RE_RDY_OUT or RE_OUT_N)
		 
     begin
	next_receive = 4'h00;

	if (state_receive[IDLE])
	begin
        if (spi_incoming_byte)
			next_receive[RECEIVE_DATA] = 1'b1;
		else
			next_receive[IDLE] = 1'b1;
	end


	if (state_receive[RECEIVE_DATA])
	  begin
	     if ((in_bit_count >= 8) & spi_clk_reg)
	       next_receive[SEND_TO_HOST] = 1'b1;
		 else if(SPI_CS_N)
		    next_receive[IDLE] = 1'b1;
	     else
	       next_receive[RECEIVE_DATA] = 1'b1;
	  end

	if (state_receive[SEND_TO_HOST]) 
	  begin
	     if (RE_RDY_OUT)
			  next_receive[RECEIVE_COMPLETE] = 1'b1;
	     else
	         next_receive[SEND_TO_HOST] = 1'b1;
	  end 	     

	  if (state_receive[RECEIVE_COMPLETE]) 
	  begin
	     if (!spi_clk_reg)
			  next_receive[IDLE] = 1'b1;
	     else
	         next_receive[RECEIVE_COMPLETE] = 1'b1;
	  end 	     

`ifdef SIM
	if (state_receive == (1 << IDLE))
	  state_name_receive = "IDLE";
	else if (state_receive == (1 << RECEIVE_DATA))
	  state_name_receive = "RECEIVE_DATA";
	else if (state_receive == (1 << SEND_TO_HOST))
	  state_name_receive = "SEND_TO_HOST";
	else if (state_receive == (1 << RECEIVE_COMPLETE))
	  state_name_receive = "RECEIVE_COMPLETE";
`endif
     end // always @ (...


   //--------------------------------------------------
   // SPI Write State Machine Transmit
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state_transmit <= 6'h00;
	     state_transmit[IDLE] <= 1'b1;
	  end
	else
	  state_transmit <= next_transmit;
     end

   always @ ( /*AUTOSENSE*/state_transmit  
         or out_bit_count or we_in_reg 
		 or spi_clk_reg or bit_count_state or out_bit_count
		 or WE_RDY_OUT )
     begin
	next_transmit = 6'h00;
	 
	if (state_transmit[IDLE_TRANSMIT])
	begin
        if (we_in_reg)
			next_transmit[WRITE_REGISTER] = 1'b1;
		else
			next_transmit[IDLE_TRANSMIT] = 1'b1;
	end

	if (state_transmit[WRITE_REGISTER])
	  begin
	     if (WE_RDY_OUT)
	       next_transmit[LOAD_SHIFT_REG] = 1'b1;
	     else 
	       next_transmit[WRITE_REGISTER] = 1'b1;
	  end

	if (state_transmit[LOAD_SHIFT_REG])
	       next_transmit[DELAY_WRITE_DATA_BYTE] = 1'b1;

	if (state_transmit[DELAY_WRITE_DATA_BYTE])
	  begin
	       next_transmit[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state_transmit[WRITE_DATA_BYTE])
	  begin
	     if (( out_bit_count >= 8 ) & spi_clk_reg)
	       next_transmit[WRITE_DATA_COMPLETE] = 1'b1;
	     else
	       next_transmit[WRITE_DATA_BYTE] = 1'b1;
	  end

	if (state_transmit[WRITE_DATA_COMPLETE])
	  begin
	     if (!we_in_reg)
	       next_transmit[IDLE_TRANSMIT] = 1'b1;
	     else
	       next_transmit[WRITE_DATA_COMPLETE] = 1'b1;
	  end

	
`ifdef SIM
	if (state_transmit == (1 << IDLE_TRANSMIT))
	  state_name_transmit = "IDLE_TRANSMIT";
	else if (state_transmit == (1 << WRITE_REGISTER))
	  state_name_transmit = "WRITE_REGISTER";
	else if (state_transmit == (1 << LOAD_SHIFT_REG))
	  state_name_transmit = "LOAD_SHIFT_REG";
	else if (state_transmit == (1 << DELAY_WRITE_DATA_BYTE))
	  state_name_transmit = "DELAY_WRITE_DATA_BYTE";
	else if (state_transmit == (1 << WRITE_DATA_BYTE))
	  state_name_transmit = "WRITE_DATA_BYTE";
	else if (state_transmit == (1 << WRITE_DATA_COMPLETE))
	  state_name_transmit = "WRITE_DATA_COMPLETE";
`endif
     end // always @ (...
   

endmodule // spi_slave

